////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty. 
//  In no event will the authors be held liable for any damages arising 
//  from the use of this software.
// 
//  Authors:
//      Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //

#ifndef ARR_H
#define ARR_H

#include <assert.h>
#include <iostream.h>

#define forall(_ARR,_EL) for (int _i=0,_EL=_ARR[0];_i<_ARR.Len();_i++,_EL=_ARR[_i]) 

template <class T>
class ARR
{
    public:
        ARR() : data(0), len(0), alloclen(0) {}

        ARR(int k) : data(0), len(0), alloclen(0) {
            SetLen(k);
        }

        ARR(int k,const T *dat) : data(0), len(0), alloclen(0) {
            Set(k,dat);
        }

        ARR(const ARR &copy) : data(0), len(0), alloclen(0) {
            (*this) = copy;
        }
            
        ~ARR() { 
            if (data!=0) { assert(alloclen); delete [] data; }
        }

        void SetLen(int i) {
            assert( len<=alloclen /* Memory corrupted if fails */ );
            assert(i>=0);
			Clear();
            if (i==0) { return; }
            if (i>alloclen) { AllocAndCopyLen(i); }
            len = i;
        }
        
        void Set(int k,const T *dat) {
            assert( len<=alloclen /* Memory corrupted if fails */ );
            assert(dat);
            SetLen(k);
            for (int i=0;i<len;i++) data[i] = dat[i];
        }
        
        int Len() const { 

            assert( len<=alloclen /* Memory corrupted if fails */ );
            return len; 
        }

        T& operator[](int i) { 
            assert( len<=alloclen /* Memory corrupted if fails */ );
            assert(i>=0); 
            assert(i<len); 
            return data[i]; 
        }

        const T &operator[](int i) const { 
            assert( len<=alloclen /* Memory corrupted if fails */ );
            assert(i>=0); 
            assert(i<len); 
            return data[i]; 
        }

        T& Tail(int k=0) { 
            assert( len<=alloclen /* Memory corrupted if fails */ );
            assert(len>k); 
            return data[len-1-k]; 
        }

        T Tail(int k=0) const { 
            assert( len<=alloclen /* Memory corrupted if fails */ );
            assert(len>k); 
            return data[len-1-k]; 
        }

        T *operator()() { return data; }
        T *addr() { return data; }
        T *addr(int i) { return &data[i]; }
        const T *addr() const { return data; }

        void operator=(const ARR<T> &copy) {
            assert( len<=alloclen /* Memory corrupted if fails */ );
            Clear();
            SetLen(copy.Len());
            memcpy(data,copy.data,sizeof(T)*len);
        }
        
        void operator=(const T &elem) {
            for (int i=0;i<Len();i++) data[i] = elem;
        }
        
        void InsertTail(T d) {
            assert( len<=alloclen /* Memory corrupted if fails */ );
            if (alloclen <= len+1) {
                AllocAndCopyLen(len*2+1);
            }
            len++;
            (*this)[len-1] = d;
        }

        void InsertHead(T d) {
            assert( len<=alloclen /* Memory corrupted if fails */ );
            if (alloclen <= len+1) {
                AllocAndCopyLen(len*2+1,1);
            } else {
                for (int i=len-1;i>=0;i--) 
                    data[i+1] = data[i];
            }
            (*this)[0] = d;
            len++;
        }

        void Clear(bool NO_DEALLOC=false) {
            assert( len<=alloclen /* Memory corrupted if fails */ );
            if (NO_DEALLOC==false) {
                if (data!=0) { assert(alloclen!=0); delete [] data; data=0; }
                alloclen = 0;
            }
            len = 0;
        }

        friend ostream& operator<<(ostream &out,const ARR<T> &arr);
        friend double Norm2(const ARR<T> &arr);
        friend T Sum(const ARR<T> &arr);
        
    private:
        int BoundsCheck(int i) const {
            return (i<0 || i>=len) ? -1 : i;
        }
        
        void AllocAndCopyLen(int L,int shift=0) {
            assert( len<=alloclen /* Memory corrupted if fails */ );
            assert(L>0);
            alloclen = L;
            T *tmp = new T[alloclen];
            assert(tmp);
            if (data!=0) {
                int min = (alloclen<len)?alloclen:len;
                for (int k=0;k<min;k++) {
                    tmp[shift+k] = data[k];
                }
//                memcpy(&tmp[shift],data,sizeof(T)*min);
                delete [] data; 
                data = 0;
            }
            data = tmp;
        }

    protected:
        T *data;
        int len;
        int alloclen;
};


template <class T>
ostream& operator<<(ostream &out,const ARR<T> &arr) {
    out.form("ARR (0x%x): len=%d, alloc=%d\n",(int)arr.data,arr.len,arr.alloclen);
    for (int i=0;i<arr.len;i++) {
        out << "    [" << i << "] = " << arr[i] << "\n";
    }
    return out;
}

template <class T>
double Norm2(const ARR<T> &arr)
{
    double sum=0;
    for (int i=0;i<arr.Len();i++) 
        sum += arr[i]*arr[i];
    return sum;
}

template <class T>
T Sum(const ARR<T> &arr)
{
    T sum=0;
    for (int i=0;i<arr.Len();i++) 
        sum += arr[i];
    return sum;
}

#endif
