////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//      Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //
#include "brush.H"
#include "XmdvTool.H"
#include <stdio.h>

Brush brushes[MAXBRUSH];        // the brushes
Brush all_brushes[MAXBRUSH];    // the complete brush set
Brush *curr_brush=0;            // the current brush

// Set initial brush position in the middle of a dimension
// and size to be 50 percent of whole range.  Also
// set initial enable/display states.
//
void InitBrushSet()
{
    for (int br=0; br<MAXBRUSH; br++)
    {
        brushes[br].num = br;
        brushes[br].pos.SetLen(dataset->all_dims);
        brushes[br].size.SetLen(dataset->all_dims);
        brushes[br].bound_offset.SetLen(dataset->all_dims);
        all_brushes[br].pos.SetLen(dataset->all_dims);
        all_brushes[br].size.SetLen(dataset->all_dims);
        all_brushes[br].bound_offset.SetLen(dataset->all_dims);

        for (int dim=0; dim<dataset->all_dims; dim++)
        {
            brushes[br].pos[dim] = (dataset->dim_max[dim] + dataset->dim_min[dim]) / 2.0;
            brushes[br].size[dim] = (dataset->dim_max[dim] - dataset->dim_min[dim]) / 2.0;
            brushes[br].bound_offset[dim] = (dataset->dim_max[dim] - dataset->dim_min[dim]) / 10.0;
            all_brushes[br].pos[dim] = brushes[br].pos[dim];
            all_brushes[br].size[dim] = brushes[br].size[dim];
            all_brushes[br].bound_offset[dim] = brushes[br].bound_offset[dim];
        }
        brushes[br].enable = false;
        brushes[br].display = false;
    }

    // Enable the first brush by default
    brushes[0].enable = true;
    brushes[0].display = true;

    // Set the current brush
    SetActiveBrush(0);
}

int SetActiveBrush(int which)
{
    if (which<0 || which>=MAXBRUSH) return 1;

    curr_brush = &brushes[which];
    return 0;
}

int SetBrushEnableState(int which, bool state)
{
    if (which<0 || which>=MAXBRUSH) return 1;

    brushes[which].enable = state;
    return 0;
}

int SetBrushDisplayState(int which, bool state)
{
    if (which<0 || which>=MAXBRUSH) return 1;

    brushes[which].display = state;
    return 0;
}

int SetBrushBoundaryType(int which, Brush::BoundaryType type)
{
    if (which<0 || which>=MAXBRUSH) return 1;

    brushes[which].bound_type = type;
    return 0;
}

// Given a data point and a brush number, this
// function checks if the point is covered by the brush
//
bool CheckBrushCoverage(const ARR<float> &point, Brush *brush)
{
    int i;
    switch(brush->bound_type)
    {
    case Brush::STEP:
        for (i=0; i<dataset->dims; i++)
            if (point[i] < brush->Min(i) || point[i] > brush->Max(i))
                return false;
        return true;

    case Brush::RAMP:
        for (i=0; i<dataset->dims; i++)
            if (point[i] < brush->RampMin(i) || point[i] > brush->RampMax(i))
                return false;
        return true;
    }
    return false;
}

// Given a data point and a brush number, this
// function checks the amount that the point
// is covered by the brush.  Only makes sense
// for brushes with non stepped boundaries.
// Returns the amount of brush coverage (0..1)
//
float PercentBrushCoverage(const ARR<float> &point, Brush *brush)
{
    switch (brush->bound_type)
    {
    case Brush::STEP: {
        for (int i=0; i<dataset->dims; i++)
        {
            if (point[i] < brush->Min(i) || point[i] > brush->Max(i))
                return 0.0;
        }
        return 1.0;
    }

    case Brush::RAMP: {
        // compute coverage as linear interpolation between brush
        // start and end
        //
        float coverage = 1e8;

        int i;
        for (i=0; i<dataset->dims; i++)
        {
            float min = brush->Min(i);
            float max = brush->Max(i);
            float offset = brush->bound_offset[i];

            float dim_coverage = 0;

            if (point[i] >= min && point[i] <= max) {
                // Point is fully in step brush
                dim_coverage = 1.0;
            } else
            if (point[i] >= (min - offset) && point[i] < min) {
                // Point is in lower ramp
                dim_coverage = (offset - (min - point[i])) / offset;
            } else
            if (point[i] <= (max + offset) && point[i] > max) {
                // Point is in upper ramp
                dim_coverage = (offset - (point[i] - max)) / offset;
            } else {
                // No coverage, return 0.0
                return 0.;
            }

            coverage = MIN(coverage,dim_coverage);
            //coverage += dim_coverage;
        }
        //coverage /= (float)dataset->dims;
        return coverage;

    }

    default:
        fprintf(stderr, "Invalid boundary type in PercentBrushCoverage()\n");
    }
    return 0;
}

// Given a data point, returns the number of brushes
// that contain that point.
//
int CoverageRank(const ARR<float> &point)
{
    int coverage=0;
    for (int br=0; br<MAXBRUSH; br++)
    {
        Brush *brush = &brushes[br];
        // skip this brush if disabled
        if (brush->enable == false) continue;
        if (CheckBrushCoverage(point, brush)) coverage++;
    }
    return coverage;
}

//
// Check if the brush currently covers this data point
//
bool Brush::IsPointCovered(const ARR<float> &point)
{
    for (int i = 0;i < dataset->dims;i++) {
        if (point[i] < Min(i) || point[i] > Max(i))
            return false;
    }
    return true;
}

//
// Automatic global brush resizing
//
void Brush::Resize(float factor)
{
    if (!dataset) return;

    if (factor == 0) {
        // Resize to max
        for (int d=0; d<dataset->dims; d++) {
            pos[d] = (dataset->dim_max[d] + dataset->dim_min[d]) / 2.0;
            size[d] = dataset->dim_max[d] - dataset->dim_min[d];
        }
    } else {
        for (int d=0; d<dataset->dims; d++) {
            size[d] *= factor;
        }
    }
}

float Brush::Min(int d)
{
    return pos[d] - size[d]*.5;
}

float Brush::Max(int d)
{
    return pos[d] + size[d]*.5;
}

float Brush::RampMin(int d)
{
    return Min(d) - bound_offset[d];
}

float Brush::RampMax(int d)
{
    return Max(d) + bound_offset[d];
}

void Brush::SetMinMax(int d,float lo,float hi)
{
    pos[d] = (hi + lo)*.5;
    size[d] = (hi - lo);
}
