////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//      Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //

#include "brush_oper.H"
#include "XmdvTool.H"
#include "cluster.H"
#include <string.h>
#include <stdio.h>

BrushOperation brush_ops[MAXOPER]; // the operations
int num_oper=0;         // number of operations
int list_edit;          // the list to edit

char *oper_names[MAXOPER] = {NULL}; // the operation names

//
// Initialize the brush operations list
//
void BrushOperListInit()
{
    // Initialize one highlight operation for the first brush
    num_oper = 1;
    brush_ops[0].num_brush = 1;
    brush_ops[0].not[0] = 0;
    brush_ops[0].brush[0] = 0;
    brush_ops[0].oper = HIGH;
/*
    num_oper = 2;
    brush_ops[0].num_brush = 1;
    brush_ops[0].not[0] = 0;
    brush_ops[0].brush[0] = 0;
    brush_ops[0].oper = HIGH;

    brush_ops[1].num_brush = 1;
    brush_ops[1].not[0] = 1;
    brush_ops[1].brush[0] = 0;
    brush_ops[1].oper = MASK;
*/
}

//
// Instantiate a new brush. Returns the new operation number.
//
int BrushOperNew()
{
    num_oper++;
    assert(num_oper<MAXOPER /* Can only support max of 30 operations */ );
    return num_oper-1;
}

//
// Delete a brush
//
void BrushOperDelete(int which)
{
    for (int i=which+1; i<num_oper; i++)
        brush_ops[i-1] = brush_ops[i];

    num_oper--;
    if (num_oper<0) num_oper=0;
}

//
// Lower priority of brush
//
void BrushOperLowerPriority(int which)
{
    if (which == num_oper-1) return;

    // change the priority
    BrushOperation tmp = brush_ops[which];
    brush_ops[which] = brush_ops[which+1];
    brush_ops[which+1] = tmp;
}

//
// Raise priority of brush
//
void BrushOperRaisePriority(int which)
{
    if (which == 0) return;

    BrushOperation tmp = brush_ops[which];
    brush_ops[which] = brush_ops[which-1];
    brush_ops[which-1] = tmp;
}

//
// Generate an ASCII name for an operation from
// an Operation structure.
//
const char *BrushOperGenerateString(int which)
{
    static char buf[512];

    BrushOperation *op = &brush_ops[which];

    // Initialize buffer
    sprintf(buf, "%2d: ", which+1);

    for (int i=0; i<op->num_brush; i++)
    {
        // Add a NOT if necessary
        if (op->not[i]) strcat(buf, "NOT ");

        // Add the brush name
        char buf2[256];
        sprintf(buf2, "B%d ", op->brush[i]+1);
        strcat(buf, buf2);

        // Add a conjunction if there are more brushes
        if (i < op->num_brush-1) {
            switch(op->conj[i])
            {
            case AND:
                strcat(buf, "AND ");
                break;
            case OR:
                strcat(buf, "OR ");
                break;
            case XOR:
                strcat(buf, "XOR ");
                break;
            }
        }
    }
    if (op->num_brush==0) strcat(buf,"<no brush> ");

    strcat(buf,": ");

    // Now add the operation(s)
    if (op->oper & HIGH)
        strcat(buf, "highlight ");

    if (op->oper & MASK)
        strcat(buf, "mask ");

    if (op->oper & VALUES)
        strcat(buf, "values ");

    if (op->oper & AVERAGE)
        strcat(buf, "average ");

    if (op->oper == 0) {
        strcat(buf,"<null operation>");
    }

    // copy the name and return it
    return buf;
}

//
// Compute if a point is covered by the given brush
// operation.
//
bool BrushOperCheckPoint(const ARR<float> &point, int which)
{
    BrushOperation *op = &brush_ops[which];

    // Check to make sure all brushes are enabled first
    int i;
    for (i=0; i<op->num_brush; i++) {
        int brush_num = op->brush[i];
        if (brushes[brush_num].enable == false)
            return false;
    }
    // Set the result to the initial brush coverage
    bool result = op->not[0] ? !CheckBrushCoverage(point, &brushes[op->brush[0]]) :
                           CheckBrushCoverage(point, &brushes[op->brush[0]]);

    for (i=1; i<op->num_brush; i++)
    {
        bool next_br = op->not[i] ? !CheckBrushCoverage(point, &brushes[op->brush[i]]) :
                                CheckBrushCoverage(point, &brushes[op->brush[i]]);

        switch(op->conj[i-1])
        {
        case AND:
            result &= next_br;
            break;
        case OR:
            result |= next_br;
            break;
        case XOR:
            result ^= next_br;
            break;
        }
    }
    return result;
}

//
// Compute the percent of coverage from a brush.  Only really makes sense for
// brushes with non stepped boundaries. Returns the coverage percent (0..1).
//
static float PercentOperationPoint(const ARR<float> &point, int which)
{
    BrushOperation *op = &brush_ops[which];

    // Check to make sure all brushes are enabled first
    int i;
    for (i=0; i<op->num_brush; i++)
        if (brushes[op->brush[i]].enable == false)
            return 0.;

    // Set the result to the initial brush coverage
    float coverage = PercentBrushCoverage(point, &brushes[op->brush[0]]);
    if (op->not[0]) {
        assert( 0 /* Not till multiops is ported */ );
        coverage = 1-coverage;
    }

    for (i=1; i<op->num_brush; i++)
    {
        assert( 0 /* Not till multiops is ported */ );
        float next_br = PercentBrushCoverage(point, &brushes[op->brush[i]]);
        if (op->not[i]) next_br = 1-next_br;

        switch(op->conj[i-1])
        {
        case AND:
            coverage = MIN(coverage, next_br);
            break;
        case OR:
            coverage = MAX(coverage, next_br);
            break;
        case XOR:
            coverage += next_br;
            if (coverage > 1.0) coverage = 2.0 - coverage;
            break;
        }
    }

    return coverage;
}

//
// Quick check to see if point is out of min/max range
//
static bool ExceedsDimensionBoundaries(const ARR<float> &data)
{
    for (int i=0;i<dataset->dims;i++) {
        if (data[i]<dataset->dim_min[i]) return true;
        if (data[i]>dataset->dim_max[i]) return true;
    }
    return false;
}

//
// This function determines the color that a point should be from the various
// brush_ops defined. The lowest numbered operation has precedence if the
// point is covered by multiple brush_ops. Returns the color cell to paint the
// point with.
//
static void BrushOperComputeColor(const ARR<float> &data, okcdata::mark_type &what_type, RGBt &what_color)
{
    // If the point is out of the dim_min/max range,
    // disable it (this happens during zooming)
    //
    if (ExceedsDimensionBoundaries(data)) {
        what_type = okcdata::DISABLED;
        return;
    }

    // The first operation has highest priority
    //
    for (int i=0; i<num_oper; i++)
    {
        if (brush_ops[i].num_brush == 0) continue;

        bool enclosed = BrushOperCheckPoint(data, i);
        if (!enclosed) continue;

        if (brush_ops[i].oper & HIGH)
        {
            what_type = okcdata::COLORED;

// NEED TO RETHINK THIS CONDITION....
            if (brush_ops[i].num_brush == 1 && brushes[brush_ops[i].brush[0]].bound_type==Brush::RAMP) {
                // Can't resolve multiple ramp brushes ?!?
                float coverage = PercentOperationPoint(data,i);
                assert(coverage > 0);
                RGBt cdata = interface_color[BG]; // [DATA]?
                RGBt chigh = interface_color[HIGHLIGHT];

                if (coverage < 1.0) coverage -= 0.5; // HACK : To make fading more obvious

                coverage = CLAMP(coverage,0.0,1.0);
                RGBt color;
                color.R = (coverage)*chigh.R + (1-coverage)*cdata.R;
                color.G = (coverage)*chigh.G + (1-coverage)*cdata.G;
                color.B = (coverage)*chigh.B + (1-coverage)*cdata.B;
                what_color = color;

                return;
            }

            what_color = interface_color[HIGHLIGHT];
            return;
        } else
        if (brush_ops[i].oper & MASK) {
            what_type = okcdata::DISABLED;
            return;
        }
    }

    // If point is not covered by any operations, just
    // paint it in the data color.
    //
    what_type = okcdata::REGULAR;
    what_color = interface_color[DATA];
}

//
// Scans all points in the dataset and assign each a color
// for drawing.
//
void BrushOperMarkDataset()
{
    ARR<float> data(dataset->dims);

    int i;
    for (i=0;i<dataset->data_size;i++) {
        dataset->GetData(data,i);
        BrushOperComputeColor(data,dataset->mark[i],dataset->color[i]);
    }

    // Traverse the cluster tree and mark the clusters that are within the brush
    if (!btree) return;

    for (i=0;i<btree->nodes.Len();i++) {
        Cluster &c = btree->nodes[i];
        RGBt dummy_color;
        BrushOperComputeColor(c.mean,c.mark,dummy_color);
    }
}

//
// Given an operation number, computes the average
// of all the points contained by the operation and
// stores the result in aver.
//
int BrushOperCompAverage(int op, ARR<float> &aver)
{
    int i;
    for (i=0; i<dataset->dims; i++) aver[i] = 0.0;

    int count=0;
    ARR<float> data(dataset->dims);

    for (i=0; i<dataset->data_size; i++)
    {
        dataset->GetData(data, i);

        if (BrushOperCheckPoint(data, op))
        {
            for (int j=0; j<dataset->dims; j++) aver[j] += data[j];
            count++;
        }
    }

    if (count!=0) {
        for (i=0; i<dataset->dims; i++) aver[i] /= count;
    }

    return count;
}
