////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//      Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //
//modified by Jing Yang, 5/2/2000


#include "cluster.H"
#include "XmdvTool.H"
#include "brush.H"
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include "cluster_hist.H"
#include "cluster_nav.H"
#include "cluster_parcoord.H"
#include "treemap.H"

ClusterTree *btree = 0;

bool ClusterTree::ReadClusterTreeFile(const char* fname)
{
    int k, i, loopi;
    assert(dataset);

    FILE *fptr = fopen(fname,"r");
    if (fptr==NULL) {
        printf("Cluster file \"%s\" does not exist. Program terminated.\n", fname);
//        abort();
        return false;
    }

    clusterfile = strdup(fname);

    // Get file size
    int num_nodes;
    if (fscanf(fptr, "%d %d\n", &num_nodes, &dimension)<=0)
    {
        printf("Error reading the first line of the file %s\n", fname);
        printf("Format of the first line: #nodes #dimension\n");
        exit(-1);
    }
    assert(dimension == dataset->all_dims);

    // Allocate space for all nodes
    //
    nodes.SetLen(num_nodes);

//  selected.SetLen(num_nodes/2+1);

    printf("Number of cluster nodes=%d\n", num_nodes);


    for (i=0; i<num_nodes; i++) {
        nodes[i].sx.SetLen(dimension);
        nodes[i].mean.SetLen(dimension);
        nodes[i].max.SetLen(dimension);
        nodes[i].min.SetLen(dimension);
      //  nodes[i].all_mean.SetLen(dimension);
        nodes[i].num_children = 0;
        nodes[i].entries = 0;
    }

    for (k=0; k<num_nodes; k++ )
        for (i=0;i<dimension; i++)
            nodes[k].sx[i] = 0.0;

/* SID */
	float max, min;
    for (k=0; k<num_nodes; k++)
    {
        int id;
        fscanf(fptr, "%d %d %d", &id, &nodes[k].parent, &nodes[k].entries);

        for (loopi=0; loopi<dimension; loopi++)
		{
            i = dataset->map_file_to_buf[loopi];
            fscanf(fptr, "%f ", &(nodes[k].sx[i]));

            // Compute the mean
            if ( nodes[k].entries > 0) {

                double sum_dim = nodes[k].sx[i] * (dataset->cg_dim_max[i]-dataset->cg_dim_min[i]) +
                                 (nodes[k].entries * dataset->cg_dim_min[i]);

                nodes[k].mean[i] = sum_dim/nodes[k].entries;
            }

        }


        fscanf(fptr, "%f ", &nodes[k].sqrt_radius);

        for (loopi=0; loopi<dimension; loopi++)
		{
            i = dataset->map_file_to_buf[loopi];
            fscanf(fptr, "%f ", &max);
			nodes[k].max[i] = max * (dataset->cg_dim_max[i]-dataset->cg_dim_min[i]) + dataset->cg_dim_min[i];
		}

        for (loopi=0; loopi<dimension; loopi++)
		{
            i = dataset->map_file_to_buf[loopi];
            fscanf(fptr, "%f ", &min);
			nodes[k].min[i] = min * (dataset->cg_dim_max[i]-dataset->cg_dim_min[i]) + dataset->cg_dim_min[i];
		}

        fscanf(fptr, "\n");

        if ( nodes[k].parent >= 0 ) {
            nodes[nodes[k].parent].child.InsertTail(k);
            nodes[nodes[k].parent].num_children++;
        }

    }

    // Init the min/max extents
    //
    min_extents.SetLen(dataset->all_dims);
    max_extents.SetLen(dataset->all_dims);

    for (int d=0; d<dataset->all_dims; d++) {
        min_extents[d] = HUGE_VAL;
        max_extents[d] = -HUGE_VAL;
    }

    return true;
}

//
// To mark the highlighted clusters based on the
// brush movement in other modes
//
void hierMarkBrushedClustersByExtents(Cluster* node)
{
    BITCLEAR(node->flags,IS_BRUSHED);

    for (int i=0; i<node->num_children; i++)
        hierMarkBrushedClustersByExtents(&(btree->nodes[node->child[i]]));

    // Reached the leaf node?
    //
    if (node->num_children == 0) {
        int in=0;
        for (int k=0; k<dataset->dims; k++) {
            if (node->mean[k] >= btree->min_extents[k] && node->mean[k] <= btree->max_extents[k]) {
                in++;
            }
        }

        if (in == dataset->dims) {
            BITSET(node->flags,IS_BRUSHED);
        }

        return;
    }

    // Propagate markers up to parents
    //
    int brushed=0;
    for (int j=0; j<node->num_children; j++) {
        if (BITTEST(btree->nodes[node->child[j]].flags,IS_BRUSHED)) {
            brushed++;
        }
    }

    // All my children are marked, mark myself too
    //
    if (brushed == node->num_children) {
        BITSET(node->flags,IS_BRUSHED);
    }
}

void hierMarkBrushedClustersByColor(Cluster* node, float L, float R)
{
    BITCLEAR(node->flags,IS_BRUSHED);

    for (int i=0; i<node->num_children; i++) {
        hierMarkBrushedClustersByColor(&(btree->nodes[node->child[i]]), L, R);
    }

    // Reached the leaf node?
    //
    if (node->num_children == 0)
    {
        if (node->color>=L && node->color<=R)
        {
            BITSET(node->flags,IS_BRUSHED);

            // Setup min/max extents
            //
            for (int d=0; d<dataset->dims; d++) {
                if (btree->min_extents[d]>node->mean[d]) btree->min_extents[d]=node->mean[d];
                if (btree->max_extents[d]<node->mean[d]) btree->max_extents[d]=node->mean[d];
            }
        }

        return;
    }

    // Propagate markers up to parents
    //
    int brushed=0;
    for (int j=0; j<node->num_children; j++) {
        if (BITTEST(btree->nodes[node->child[j]].flags,IS_BRUSHED)) {
            brushed++;
        }
    }

    // If all my children are marked, mark myself too
    //
    if (brushed == node->num_children) {
        BITSET(node->flags,IS_BRUSHED);
    }
}

void hierModifyBrushParametersByExtents()
{
    for (int i=0; i<dataset->dims; i++) {
        curr_brush->size[i] = btree->max_extents[i] - btree->min_extents[i];
        curr_brush->pos[i] = btree->min_extents[i] + (curr_brush->size[i]/2.);
    }
}

void hierModifyExtentsByBrushParameters()
{
    for (int i=0; i<dataset->dims; i++) {
        btree->min_extents[i] = curr_brush->Min(i);
        btree->max_extents[i] = curr_brush->Max(i);
    }
}

void hierInitClusterColors(Cluster* node, float L, float R)
{
    node->color = .5*(L+R);

    float k = (R-L)/(float)(node->num_children);

    for (int i=0; i<node->num_children; i++) {
        hierInitClusterColors(&btree->nodes[node->child[i]], L+i*k, L+(i+1)*k);
    }
}

void hierInitClusterColorsWithDistortion(Cluster* node, float L, float R)
{
    node->color = .5*(L+R);

    float color_l=L;
    float color_r=L;

    for (int i=0; i<node->num_children; i++)
    {
        float k = (R-L) * btree->nodes[node->child[i]].entries/(float)node->entries;
        color_l = color_r;
        color_r = color_l + k;
        hierInitClusterColors(&btree->nodes[node->child[i]], color_l, color_r);
    }
}

void hierDeleteClusterTree()
{
    delete btree;
    btree = 0;
}

void hierInitClusterTree()
{
    if (btree) return;

    btree = new ClusterTree;

    int N = strlen(dataset->datafile);
//    char str[N+1];
//jing
	char* str = new char[N+1];
//jing
    strcpy(str,dataset->datafile);
//jing    strcpy(&str[N-4],".cf");
    strcpy(&str[N-4],".cg");

    if (!btree->ReadClusterTreeFile(str)) {
        delete btree;
        btree=0;
        return;
    }

    hierInitClusterColors(&btree->nodes[0], 0, 1);
    hierClearNodeFlags();
    hierModifyExtentsByBrushParameters();
    hierMarkBrushedClustersByExtents(&btree->nodes[0]);

    cfHistoInit();
    cfNavInit();
    cfTreeMapInit();

    float root_radius = btree->nodes[0].sqrt_radius;
    hierMarkContourNodes( root_radius*cluster_radius, IS_CONTOUR0 );
    hierMarkContourNodes( root_radius*brushed_radius, IS_CONTOUR1 );
	//jing
	SetupBrushParameters();
}

void hierClearNodeFlags(int bits)
{
    for (int i=0;i<btree->nodes.Len();i++) {
        Cluster &c = btree->nodes[i];
        BITCLEAR(c.flags,bits);
    }
}

void hierMarkContourNodes(float contour_radius, int FLAG)
{
    // A node c is on contour if:
    // c.parent.radius > contour_radius AND
    //  (a)  c.radius <= contour_radius, OR
    //  (b)  c.radius > contour_radius AND c is a leaf
    //
    for (int i=0;i<btree->nodes.Len();i++)
    {
        Cluster &c = btree->nodes[i];

        BITCLEAR(c.flags,FLAG);

        float par_radius = HUGE_VAL;
        if (c.parent>=0) par_radius = btree->nodes[c.parent].sqrt_radius;

        if (par_radius > contour_radius)
        {
            if (c.sqrt_radius <= contour_radius || c.num_children==0)
            {
                assert(c.color<=1.0 && c.color>=0);
                BITSET(c.flags,FLAG);
            }
        }
    }
}

void hierInitExtents()
{
    for (int d=0; d<dataset->all_dims; d++) {
        btree->min_extents[d] = HUGE_VAL;
        btree->max_extents[d] = -HUGE_VAL;
    }
}


bool ClusterTree::ReadClusterTreeFile2()
{
    int k, i, loopi;
    assert(dataset);

    FILE *fptr = fopen(clusterfile,"r");
    if (fptr==NULL) {
        printf("Cluster file \"%s\" does not exist. Program terminated.\n", clusterfile);
//        abort();
        return false;
    }

    // Get file size
    int num_nodes;
    if (fscanf(fptr, "%d %d\n", &num_nodes, &dimension)<=0)
    {
        printf("Error reading the first line of the file %s\n", clusterfile);
        printf("Format of the first line: #nodes #dimension\n");
        exit(-1);
    }
    assert(dimension == dataset->all_dims);

    // Allocate space for all nodes
    //
	float max, min, tempf;
    int tempi;
	for (k=0; k<num_nodes; k++)
    {
        fscanf(fptr, "%d %d %d", &tempi, &tempi, &tempi);

        for (loopi=0; loopi<dimension; loopi++)
		{
            i = dataset->map_file_to_buf[loopi];
            fscanf(fptr, "%f ", &tempf);

            // Compute the mean
            if ( nodes[k].entries > 0) {

                double sum_dim = nodes[k].sx[i] * (dataset->cg_dim_max[i]-dataset->cg_dim_min[i]) +
                                 (nodes[k].entries * dataset->cg_dim_min[i]);

                nodes[k].mean[i] = sum_dim/nodes[k].entries;
            }

        }


        fscanf(fptr, "%f ", &tempf);

        for (loopi=0; loopi<dimension; loopi++)
		{
            i = dataset->map_file_to_buf[loopi];
            fscanf(fptr, "%f ", &max);
			nodes[k].max[i] = max * (dataset->cg_dim_max[i]-dataset->cg_dim_min[i]) + dataset->cg_dim_min[i];
		}

        for (loopi=0; loopi<dimension; loopi++)
		{
            i = dataset->map_file_to_buf[loopi];
            fscanf(fptr, "%f ", &min);
			nodes[k].min[i] = min * (dataset->cg_dim_max[i]-dataset->cg_dim_min[i]) + dataset->cg_dim_min[i];
		}

        fscanf(fptr, "\n");
    }

    return true;
}


void hierSetMeanOnOff()
{
	btree->ReadClusterTreeFile2();
    for (int b=0;b<btree->nodes.Len();b++)
    {
        Cluster &c = btree->nodes[b];
        int dims=0;
        for (int i=0;i<dataset->all_dims;i++)
            if (dataset->on[i]) {
                c.mean[dims] = c.mean[i];
                c.max[dims] = c.max[i];
                c.min[dims] = c.min[i];
                dims++;
            }
    }
}


bool ClusterTree::ReadClusterTreeFile3()
{
    int k, i, loopi;
    assert(dataset);

    FILE *fptr = fopen(clusterfile,"r");
    if (fptr==NULL) {
        printf("Cluster file \"%s\" does not exist. Program terminated.\n", clusterfile);
//        abort();
        return false;
    }

    // Get file size
    int num_nodes;
    if (fscanf(fptr, "%d %d\n", &num_nodes, &dimension)<=0)
    {
        printf("Error reading the first line of the file %s\n", clusterfile);
        printf("Format of the first line: #nodes #dimension\n");
        exit(-1);
    }
    assert(dimension == dataset->all_dims);

    // Allocate space for all nodes
    //
	float max, min, tempf;
    int tempi;
	for (k=0; k<num_nodes; k++)
    {
        fscanf(fptr, "%d %d %d", &tempi, &tempi, &tempi);

        for (loopi=0; loopi<dimension; loopi++)
		{
            i = dataset->map_file_to_buf[loopi];
            fscanf(fptr, "%f ", &(nodes[k].sx[i]));

            // Compute the mean
            if ( nodes[k].entries > 0) {

                double sum_dim = nodes[k].sx[i] * (dataset->cg_dim_max[i]-dataset->cg_dim_min[i]) +
                                 (nodes[k].entries * dataset->cg_dim_min[i]);

                nodes[k].mean[i] = sum_dim/nodes[k].entries;
            }

        }


        fscanf(fptr, "%f ", &tempf);

        for (loopi=0; loopi<dimension; loopi++)
		{
            i = dataset->map_file_to_buf[loopi];
			fscanf(fptr, "%f ", &max);
			nodes[k].max[i] = max * (dataset->cg_dim_max[i]-dataset->cg_dim_min[i]) + dataset->cg_dim_min[i];
		}

        for (loopi=0; loopi<dimension; loopi++)
		{
            i = dataset->map_file_to_buf[loopi];
            fscanf(fptr, "%f ", &min);
			nodes[k].min[i] = min * (dataset->cg_dim_max[i]-dataset->cg_dim_min[i]) + dataset->cg_dim_min[i];
		}

        fscanf(fptr, "\n");
    }

    return true;
}


void hierReorderDimension()
{
	btree->ReadClusterTreeFile3();
    for (int b=0;b<btree->nodes.Len();b++)
    {
        Cluster &c = btree->nodes[b];
        int dims=0;
        for (int i=0;i<dataset->all_dims;i++)
            if (dataset->on[i]) {
                c.mean[dims] = c.mean[i];
                c.max[dims] = c.max[i];
                c.min[dims] = c.min[i];
                dims++;
            }
    }
}
