////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//      Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //

#include "cluster_hist.H"
#include "cluster_nav.H"
#include "cluster_parcoord.H"
#include "cluster.H"
#include "ogl.H"
#include "XmdvTool.H"

#define HISTO_RES 100
#define BORDER 5
#define TOP_BORDER BORDER
#define BOT_BORDER 3*BORDER

static double max_radius=0.1;
ARR<int> histo;
static int max_nodes=0;
static float line_ht = 0;

void RegisterCfHistCanvas(grCanvas *canvas)
{
    assert(canvas);
    canvas->display_proc = cfHistoRedrawCanvas;
    canvas->motion_proc = cfHistoMotionNotify;
    canvas->buttonpress_proc = cfHistoButtonPress;
}

//
//  Traverse the tree up to the radius indicated
//  and return the number of nodes selected
//
static void nodes_selected(Cluster* node)
{
    max_radius = btree->nodes[0].sqrt_radius;
    int pos = (int)((node->sqrt_radius/max_radius)*HISTO_RES);

    if (pos>=HISTO_RES) pos=HISTO_RES-1;

    histo[pos]++;

    for (int i=0; i<node->num_children; i++)
        nodes_selected(&btree->nodes[node->child[i]]);

    return;
}

void cfHistoInit()
{
    assert(btree);

    histo.SetLen(HISTO_RES);

    int i;
    for (i=0; i<HISTO_RES; i++)
        histo[i]=0;

    nodes_selected(&btree->nodes[0]);

    max_nodes=0;
    for (i=1; i<HISTO_RES; i++ ) {
        // Get the maximum number of nodes
        if (histo[i] > max_nodes) max_nodes=histo[i];
    }

    line_ht = max_nodes*0.8;
}

void cfHistoRedrawCanvas( grCanvas *canvas )
{
    if (btree==0) return;   // btree not built yet

    //
    // Setup the histogram canvas
    //
    canvas->MakeCurrent();

    //
    // Setup window parameters.
    //
    int W, H;
    canvas->GetSize(&W,&H);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glViewport( BORDER, BOT_BORDER, W-2*BORDER, H-BOT_BORDER-TOP_BORDER);
    glOrtho(0,HISTO_RES,0,max_nodes,-1,1);

    canvas->SetBackground(0x000000);
    canvas->Clear();
    canvas->SetCursor("sb_h_double_arrow");

    //
    // Draw the histogram
    //
    canvas->SetForeground(interface_color[DATA]);
    for (int i=0; i<HISTO_RES; i++) {
        canvas->DrawLine(i, 0, i, histo[i]);
    }

    //
    // Indicator for unbrushed clusters
    //
    int outer_rad_pos = (int)(cluster_radius * HISTO_RES);
    canvas->SetForeground(interface_color[STAT1]);
    canvas->DrawLine(outer_rad_pos, 0, outer_rad_pos, line_ht);

    //
    // Indicator for brushed clusters
    //
    int inner_rad_pos = (int)(brushed_radius*HISTO_RES);
    canvas->SetForeground(interface_color[HIGHLIGHT]);
    canvas->DrawLine(inner_rad_pos, 0, inner_rad_pos, line_ht*0.5);

    char str[10];

    // Write the unbrushed radius value
    //
    sprintf(str,"%.02f", cluster_radius);
    canvas->SetForeground(0xffffff);
    canvas->DrawString(outer_rad_pos+2,max_nodes*0.85,str);

    // Write the brushed radius value
    //
    glViewport( BORDER, BORDER, W-2*BORDER, 2*BORDER);
    sprintf(str,"%.02f", brushed_radius);
    canvas->DrawString(inner_rad_pos+2,0,str);
}

static int MOVE_WHICH = -1;

void cfHistoMotionNotify(grCanvas *canvas, int x, int y)
{
    if (btree==0) return;
    if (canvas->button[0] != true) return;

    int W,H;
    canvas->GetSize(&W,&H);

    //
    // Map x mouse coordinates to histogram coordinates [0,1]
    //
    float xx = (x-BORDER)/(float)(W-2*BORDER);

    if (xx<0) xx = 0;
    else if (xx>1) xx = 1;

    //
    // Map y mouse coordinates to histogram coordinates [0,1]
    //
//  float yy = ((y-BORDER)/(float)(H-2*BORDER)) * line_ht;
    float yy = ((y-TOP_BORDER)/(float)(H-BOT_BORDER-TOP_BORDER)) * line_ht;

    //
    // Determine to move inner or outer radius indicator
    //
    if (MOVE_WHICH < 0) {
        float mid = line_ht*0.5;
        if (yy>mid) {
            MOVE_WHICH = 0;
        } else {
            MOVE_WHICH = 1;
        }
    }

    if (canvas->shiftkey) {
        if (MOVE_WHICH == 0) {
            cluster_radius = xx;
        } else {
            brushed_radius = xx;
        }
    } else {
//      SetRadiusSlider( xx );
        cluster_radius = brushed_radius = xx;
    }

/* SID */
    float root_radius = btree->nodes[0].sqrt_radius;
    hierMarkContourNodes( root_radius*cluster_radius, IS_CONTOUR0 );
    hierMarkContourNodes( root_radius*brushed_radius, IS_CONTOUR1 );
    appUpdateClusterRadiusSliders(Togl_Interp(canvas->GetWidget()));

    appRefreshAllCanvas(Togl_Interp(canvas->GetWidget()));
}

void cfHistoButtonPress(grCanvas *canvas, int button, int x, int y)
{
    if (btree==0) return;

    MOVE_WHICH = -1;
    cfHistoMotionNotify(canvas,x,y);
}
