////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//      Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //
#include "cluster_nav.H"
#include "cluster_parcoord.H"
#include "cluster_hist.H"
#include "cluster.H"
#include "ARR.H"
#include "ogl.H"
#include "colormap.H"
#include "XmdvTool.H"
#include "brush_oper.H"
#include <math.h>
#include <stdlib.h>

#define CONTOUR_RES 25

static ARR<float> contour_radius;
static ARR< ARR<float> > tree_contour;
static ARR< ARR<float> > tree_color;
static int max_display_height;

void RegisterCfNavCanvas(grCanvas *canvas)
{
//Jing
	assert(canvas);
//jing
    canvas->display_proc = cfNavRedrawCanvas;
    canvas->motion_proc = cfNavMotionNotify;
    canvas->buttonpress_proc = cfNavButtonPress;
    canvas->buttonrelease_proc = cfNavButtonRelease;
}

static void FindTreeHeight(Cluster* node, int curr_ht, int *max_ht)
{
    if (curr_ht>*max_ht) *max_ht = curr_ht;

    for (int i=0; i<node->num_children; i++)
        FindTreeHeight(&btree->nodes[node->child[i]], curr_ht+1, max_ht);
}

static void FindRadiusDistribution()
{
    assert(histo.Len()!=0);

    int MIN_GAP = 5;

    float R = btree->nodes[0].sqrt_radius;

    contour_radius.SetLen(0);

    for (int i=histo.Len()-1, previ=1000000;i>=0;i--)
    {
        if (histo[i]==0) continue;
        if ((previ-i)<MIN_GAP) continue;

        contour_radius.InsertTail( R*i/float(histo.Len()-1) );
        previ = i;
    }
    contour_radius.InsertTail(0.0);
}

static void FindTreeContour(Cluster *node,int curr_ht,float rad,ARR<float> &contour,ARR<float> &color,float L,float R)
{
    if (node->parent != -1)
    {
        Cluster *parent = &btree->nodes[node->parent];

        //
        // If it's a clustered node and a leaf
        //
        if (node->num_children==0 && node->sqrt_radius >= rad) {
            rad = node->sqrt_radius+ 0.01; // (1.0/(float)CONTOUR_RES);
        }

        if (node->sqrt_radius < rad)
        {
            //
            // Iso-radius contour cuts between node and parent.
            // "interp" is the intersection ratio between node and parent.
            // "y" is the intersection height of the contour.
            //

            float interp;

            if (parent->sqrt_radius - node->sqrt_radius < 0.0000001) interp = 0;
            else
                interp = (rad - node->sqrt_radius) /
                           (parent->sqrt_radius - node->sqrt_radius);

            float y = curr_ht - interp;

            //
            // Find the contour segment(s) this node covers: [x0,x1]
            //
            float x0 = L*CONTOUR_RES;
            float x1 = R*CONTOUR_RES;

            int xi0 = (int) ceil(x0+.000000001);
            int xi1 = (int) floor(x1-.00000001);

            for (int i=xi0;i<xi1;i++) {
                // Coverage that fills whole segments
                assert(contour[i] == 0.0);
                contour[i] = y;
                color[i] = (L+R)*.5;
            }

            if (xi0>0 && x1>xi0) {
                // Left end of the coverage
                float coverage = (xi0-x0);
                contour[xi0-1] += coverage*y;
                if (coverage>.9) color[xi0-1] = (L+R)*.5;
            }
            if (xi1<CONTOUR_RES && x0<xi1) {
                // Right end of the coverage
                float coverage = (x1-xi1);
                contour[xi1] += coverage*y;
                if (coverage>.9) color[xi1] = (L+R)*.5;
            }
            if (xi0>=x1 && xi1<=x0) {
                // Partial coverage
                float coverage = (x1-x0);
                contour[xi1] += coverage*y;
                if (coverage>.9) color[xi1] = (L+R)*.5;
            }

            return;
        }
    }

    float k = (R-L)/(float)(node->num_children);

    for (int i=0; i<node->num_children; i++) {
        FindTreeContour(&btree->nodes[node->child[i]], curr_ht+1, rad, contour, color, L+i*k, L+(i+1)*k);
    }
}

void SetupBrushParameters()
{
//    EXTENT_BRUSH = 0;
    hierInitExtents();
    hierClearNodeFlags(IS_BRUSHED);
    hierMarkBrushedClustersByColor(&btree->nodes[0], tree_brush[0], tree_brush[1]);

//Mask out since data brush and sbb is treated as separate entities
//    hierModifyBrushParametersByExtents();
//    BrushOperMarkDataset();
}

void cfNavInit()
{
    FindRadiusDistribution();

    int i;

    tree_height = 0;

    // Clear contour array
    //
    for (i=0;i<tree_contour.Len();i++) {
        tree_contour[i].Clear();
    }
    tree_contour.Clear();


    // Compute the contours using the average radii
    //
    tree_contour.SetLen(contour_radius.Len());
    tree_color.SetLen(contour_radius.Len());

    for (i=0;i<tree_contour.Len();i++) {
        tree_contour[i].SetLen(CONTOUR_RES);
        tree_color[i].SetLen(CONTOUR_RES);
        for (int j=0;j<CONTOUR_RES;j++) {
            tree_contour[i][j] = 0.0;
            tree_color[i][j] = -1.0;
        }
    }

    for (i=0;i<tree_contour.Len();i++) {
        FindTreeContour(&btree->nodes[0],0,contour_radius[i],tree_contour[i],tree_color[i],0,1);
    }

    //
    // Find the maximum depth of the contours, for display
    // purposes. The maximum depth may not be the maximum
    // height of the tree, because the contours are approximated.
    //
    max_display_height = 0;
    for (int j=0;j<CONTOUR_RES;j++) {
        if (tree_contour[tree_contour.Len()-1][j] > max_display_height)
            max_display_height = (int) ceil(tree_contour[tree_contour.Len()-1][j]);
    }
}


static void drawContours(float radius, float L, float R, float width=1.0)
{
    //
    // Find the two bounding contour radii
    //
    int i;
    for (i=0;i<contour_radius.Len();i++) {
        if (radius > contour_radius[i]) break;
    }
    if (i==0) i++;
    else if (i==contour_radius.Len()) i--;

    //
    // Interpolate between the two contours to
    // get the approximated contour for the current
    // radius. Show colormap along the contour.
    //

    glLineWidth(width);
    glBegin(GL_LINE_STRIP);

    float y_interp = (radius - contour_radius[i]) / (contour_radius[i-1]-contour_radius[i]);
    float x_gap = 1.0 / (CONTOUR_RES-1);

    for (int j=0;j<CONTOUR_RES;j++)
    {
        float y_hi = tree_contour[i][j];
        float y_lo = tree_contour[i-1][j];

        // x_interp is normalized between [0,1]
        float x_interp = j*x_gap;

        // Limit contour drawing to between L and R
        if (x_interp<L) {
            if ((x_interp+x_gap)<L) continue;
            float k = (L-x_interp)/x_gap;

            if (j<CONTOUR_RES-1) {
                y_hi = (1-k)*(tree_contour[i][j]) + k*tree_contour[i][j+1];
                y_lo = (1-k)*(tree_contour[i-1][j]) + k*tree_contour[i-1][j+1];
                x_interp = L;
            }
        }
        else
        if (x_interp>R) {
            if ((x_interp-x_gap)>R) continue;
            float k = (x_interp-R)/x_gap;

            if (j>0) {
                y_hi = (1-k)*(tree_contour[i][j]) + k*tree_contour[i][j-1];
                y_lo = (1-k)*(tree_contour[i-1][j]) + k*tree_contour[i-1][j-1];
                x_interp = R;
            }
        }

        float y = (1-y_interp)*y_hi + (y_interp)*y_lo;

        // xx is normalized between [-1,1]
        float xx = 2*x_interp - 1;

        // Now, normalize xx to trianglular coordinates
        xx *= y/(float)max_display_height;

        if (tree_color[i][j] != -1) { x_interp = tree_color[i][j]; }
        unsigned long color = CMAP[ int(x_interp*(CMAPSIZE-1)) ];
        glColor3ub((color>>16)&0xff,(color>>8)&0xff,color&0xff);

        glVertex2f(xx,y);
    }

    glEnd();
    glLineWidth(1.0);
}

#define BORDER 5
#define TOP_BORDER BORDER
#define BOT_BORDER 5*BORDER

void cfNavRedrawCanvas( grCanvas *canvas )
{
    if (!dataset || !btree) return;

    canvas->MakeCurrent();

    //
    // Setup window parameters.
    //
    int W,H;
    canvas->GetSize(&W,&H);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glViewport( BORDER, BOT_BORDER, W-2*BORDER, H-BOT_BORDER-TOP_BORDER);
    glOrtho(-1,+1,max_display_height+.001,0,-1,1);

    canvas->SetForeground(0xaaaaaa);
    canvas->SetBackground(0x000000);
    canvas->Clear();

    //
    // Draw the brush and brush handles
    //
    glColor3f(.5,.1,.3);
/*
    static unsigned char *stipple = 0;
    if (stipple == 0) {
        stipple = new unsigned char[128];
        for (int i=0; i<128; i++) {
            stipple[i] = rand()%256;
        }
    }

    glPolygonStipple(stipple);
    if (EXTENT_BRUSH) glEnable(GL_POLYGON_STIPPLE);
*/
    glBegin(GL_POLYGON);
    glVertex2f(0, 0);
    glVertex2f(2*tree_brush[0]-1, max_display_height);
    glVertex2f(2*tree_brush[1]-1, max_display_height);
    glEnd();

//    glDisable(GL_POLYGON_STIPPLE);

    glColor3f(.8,.8,.8);
    glPointSize(5.0);
    glBegin(GL_POINTS);
    glVertex2f(2*tree_brush[0]-1, max_display_height);
    glVertex2f(2*tree_brush[1]-1, max_display_height);
    glEnd();
    glPointSize(1.0);


    //
    // Draw the outer triangle
    //
    glColor3f(.8,.8,0);
    glBegin(GL_LINE_LOOP);
    glVertex2f(0, 0);
    glVertex2f(+1, max_display_height);
    glVertex2f(-1, max_display_height);
    glEnd();

    //
    // Draw the contours
    //
    glColor3f(.8,.8,.8);
    int i;
    for (i=0;i<tree_contour.Len();i++) {
i=tree_contour.Len()-1; // only draw the last contour
        glBegin(GL_LINE_STRIP);
        for (int j=0;j<CONTOUR_RES;j++) {
            float y = tree_contour[i][j];
            float x = ( 2*j/(float)(CONTOUR_RES-1) - 1 ) * y/(float)max_display_height;
//          printf("Vertex x[%d] = %f y=%f \n", j,x,y);
            glVertex2f(x,y);
        }
        glEnd();
    }

    //
    // Draw the current contour
    //
    float r = btree->nodes[0].sqrt_radius;
    drawContours(r*cluster_radius,0.,tree_brush[0], 2.0);
    drawContours(r*brushed_radius,tree_brush[0], tree_brush[1], 3.0);
    drawContours(r*cluster_radius,tree_brush[1], 1.0, 2.0);

    //
    // Draw the labels: tree depth, current depth
    //
    if (tree_height == 0) FindTreeHeight(&btree->nodes[0],0,&tree_height);
    char str[30];
    sprintf(str,"Tree Depth:%d", tree_height);
    canvas->SetForeground(0xffffff);
    canvas->DrawString(-1,0.5,str);

    glViewport( BORDER, BORDER, W-2*BORDER, 2*BORDER);
    CMAP.DrawColorBar();
}


enum MOUSE_MODE {
    INACTIVE,
    LEFT_BRUSH_HANDLE,
    RIGHT_BRUSH_HANDLE,
    MIDDLE_BRUSH,
    INNER_CONTOUR,
    OUTER_CONTOUR
};

static MOUSE_MODE near_mouse_mode = INACTIVE;
static MOUSE_MODE mouse_mode = INACTIVE;

static int last_pick_x;

void cfNavMotionNotify(grCanvas *canvas, int x, int y)
{
    if (!dataset || !btree) return;

    int W,H;
    canvas->GetSize(&W,&H);

    //
    // Map mouse coordinates to triangular coordinates [-1,1], [0,1]
    //
    float xx = 2*(x-BORDER)/(float)(W-2*BORDER) - 1;
    float yy = 1.0 - (y-BOT_BORDER)/(float)(H-TOP_BORDER-BOT_BORDER);
    if (yy!=0) xx /= yy;

    //
    // Check if we are picking on the brush handles
    //
    #define NEAR 0.05
    near_mouse_mode = INACTIVE;

    if (mouse_mode == INACTIVE) {

        float L = 2*tree_brush[0]-1;
        float R = 2*tree_brush[1]-1;

        if (fabs(yy-1)<NEAR) {
            if ( xx>L && xx<R ) {

                canvas->SetCursor("sb_h_double_arrow");
                near_mouse_mode = MIDDLE_BRUSH;
            } else
            if ( fabs(xx-L) < NEAR ) {

                canvas->SetCursor("right_side");
                near_mouse_mode = LEFT_BRUSH_HANDLE;
            } else
            if ( fabs(xx-R) < NEAR ) {
                canvas->SetCursor("left_side");
                near_mouse_mode = RIGHT_BRUSH_HANDLE;
            }
        } else {
            if ( xx<-1 || xx>1 || yy<0 || yy>1) {

                canvas->SetCursor("cross");
            } else
            if ( xx>L && xx< R ) {
                near_mouse_mode = INNER_CONTOUR;
                canvas->SetCursor("sb_v_double_arrow");
            } else {
                near_mouse_mode = OUTER_CONTOUR;
                canvas->SetCursor("double_arrow");
            }
        }
    }

    if (canvas->button[0] != true) return;

    switch(mouse_mode) {
    case LEFT_BRUSH_HANDLE:
        tree_brush[0] = (xx+1)*.5;
        if (tree_brush[0]<0) tree_brush[0]=0;
        if (tree_brush[0]>tree_brush[1]) tree_brush[0]=tree_brush[1];

        SetupBrushParameters();
        appRefreshAllCanvas(Togl_Interp(canvas->GetWidget()));

        return;

    case RIGHT_BRUSH_HANDLE:
        tree_brush[1] = (xx+1)*.5;
        if (tree_brush[1]>1) tree_brush[1]=1;
        if (tree_brush[1]<tree_brush[0]) tree_brush[1]=tree_brush[0];

        SetupBrushParameters();
        appRefreshAllCanvas(Togl_Interp(canvas->GetWidget()));
        return;

    case MIDDLE_BRUSH: {

        float prev_x = 2*(last_pick_x-BORDER)/(float)(W-2*BORDER) - 1;
        if (yy!=0) prev_x /= yy;

        float delta_x = (xx - prev_x)/2.0;

        tree_brush[0] += delta_x;
        tree_brush[1] += delta_x;

        if (tree_brush[1]>1) {
            delta_x = tree_brush[1]-1;
            tree_brush[0] -= delta_x;
            tree_brush[1] -= delta_x;
        } else
        if (tree_brush[0]<0) {
            delta_x = tree_brush[0];
            tree_brush[0] -= delta_x;
            tree_brush[1] -= delta_x;
        }

        last_pick_x = x;

        SetupBrushParameters();
        appRefreshAllCanvas(Togl_Interp(canvas->GetWidget()));
        return;
        }

    default:
        break;
    }

    //
    // Reject if picking outside triangle
    //
    if (xx<-1 || xx>1 || yy<0 || yy>1) return;

    //
    // Map to contour coordinates [0,CONTOUR_RES], [0,max_display_height]
    //
    yy *= max_display_height;

    xx = ((xx + 1)/2.0) * (CONTOUR_RES-1);

    int ixx = (int) xx;
    float ratio = xx-ixx;

    //
    // Find out which contour bounds "yy" at position "xx"
    //
    float cy = 0, prev_cy;
    int i;
    for (i=0;i<tree_contour.Len();i++) {
        prev_cy = cy;
        cy = tree_contour[i][ixx];
        if (ixx<CONTOUR_RES-1) cy += ratio*(tree_contour[i][ixx+1] - tree_contour[i][ixx]);
        if (cy > yy) break;
    }

    float radius;
    if (i==tree_contour.Len()) {
        radius = 0;
    } else if (i==0){
        radius = btree->nodes[0].sqrt_radius;
    } else{
        ratio = (yy-prev_cy)/(cy-prev_cy);
        radius = (1-ratio)*contour_radius[i-1] + (ratio)*contour_radius[i];
    }

    float root_radius = btree->nodes[0].sqrt_radius;
    radius /= root_radius;

    switch(mouse_mode)
    {
        case INNER_CONTOUR:
            brushed_radius = radius;
            hierMarkContourNodes( root_radius*brushed_radius, IS_CONTOUR1 );
            appUpdateClusterRadiusSliders(Togl_Interp(canvas->GetWidget()));
            break;

        case OUTER_CONTOUR:
            cluster_radius = radius;
            hierMarkContourNodes( root_radius*cluster_radius, IS_CONTOUR0 );
            appUpdateClusterRadiusSliders(Togl_Interp(canvas->GetWidget()));
            break;
        default:
            break;
    }

    appRefreshAllCanvas(Togl_Interp(canvas->GetWidget()));
}

static int prev_normal_style, prev_brush_style;

void cfNavButtonPress(grCanvas *canvas, int button, int x, int y)
{
    if (near_mouse_mode!=INACTIVE) mouse_mode = near_mouse_mode;
    last_pick_x = x;

    prev_normal_style = normal_style;
    prev_brush_style = highlight_style;

//??    normal_style |= DIM;

    cfNavMotionNotify(canvas, x, y);
}

void cfNavButtonRelease(grCanvas *canvas, int button, int x, int y)
{
    normal_style = prev_normal_style;
    highlight_style = prev_brush_style;

    near_mouse_mode = mouse_mode = INACTIVE;
    cfNavMotionNotify(canvas, x, y);
}
