////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//      Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //

#include "cluster_scatter.H"
#include "cluster_parcoord.H"
#include "ogl.H"
#include "cluster.H"
#include "XmdvTool.H"
#include "colormap.H"
#include "scatter.H"
#include "magnify.H"
#include "brush_oper.H"

// Various coordinate mappings between screen space
// and normalized data space
#define XMAP(DIM,VAL) (DIM*dataset->one_over_dims+(((VAL) - dataset->dim_min[DIM]) * dataset->one_over_dims / (dataset->dim_max[DIM] - dataset->dim_min[DIM])))
#define YMAP(DIM,VAL) (DIM*dataset->one_over_dims+((dataset->dim_max[DIM] - (VAL)) * dataset->one_over_dims / (dataset->dim_max[DIM] - dataset->dim_min[DIM])))

#define XINVMAP(DIM,VAL) (((VAL)-(DIM)*dataset->one_over_dims) * (dataset->dim_max[DIM] - dataset->dim_min[DIM]) * dataset->dims + dataset->dim_min[DIM])
#define YINVMAP(DIM,VAL) (dataset->dim_max[DIM] - ((VAL)-(DIM)*dataset->one_over_dims) * (dataset->dim_max[DIM] - dataset->dim_min[DIM]) * dataset->dims)

void RegisterHierScatterCanvas(grCanvas *canvas)
{
    canvas->display_proc = do_cluster_scatter;
    canvas->reshape_proc = grCanvas::grResizeMainCanvas;
    canvas->motion_proc = InputHierScatterMouseMotion;
    canvas->buttonpress_proc = InputHierScatterButtonPress;
    canvas->buttonrelease_proc = InputHierScatterButtonRelease;
//    canvas->keypress_proc = appCanvasInputKeyPress;
    canvas->keyrelease_proc = InputHierScatterKeyRelease;
}

//draw a brush
void DrawScatterBrush(grCanvas *canvas, Brush *brush)
{
    canvas->SetForeground( interface_color[BRUSH1+brush->num]);
    canvas->SetXorMode(false);

    for (int j=0; j<dataset->dims; j++) {
        for(int k=0; k<dataset->dims; k++)
        {
            Vec2 p, q;

            p.X = brush->Min(j);
            p.Y = brush->Max(k);

            p.X = CLAMP(p.X, dataset->dim_min[j], dataset->dim_max[j]);
            p.Y = CLAMP(p.Y, dataset->dim_min[k], dataset->dim_max[k]);

            q.X = brush->Max(j);
            q.Y = brush->Min(k);

            q.X = CLAMP(q.X, dataset->dim_min[j], dataset->dim_max[j]);
            q.Y = CLAMP(q.Y, dataset->dim_min[k], dataset->dim_max[k]);

            p.X = XMAP(j, p.X);
            p.Y = YMAP(k, p.Y);
            q.X = XMAP(j, q.X);
            q.Y = YMAP(k, q.Y);

            canvas->FillRectangle(p.X, p.Y, q.X-p.X, q.Y-p.Y);
        }
    }

    // draw boundary coverage if necessary
    if (brush->bound_type != Brush::STEP)
        DrawScatterBrushBoundRamp(canvas,brush,interface_color[BRUSH1+brush->num],interface_color[BG]);
}

//draw brushes
void DrawScatterBrushes(grCanvas *canvas)
{
    for (int brnum=0; brnum<MAXBRUSH; brnum++) {
        Brush *brush = &brushes[brnum];

        // if this brush is not being displayed, skip it
        if (brush->display == false) continue;

        DrawScatterBrush(canvas,brush);
    }
}

// draw the grids and labels of the plots
void DrawScatterGridAndLabel(grCanvas *canvas)
{
    float boxwidth = 1.0/dataset->dims;
    float boxheight = 1.0/dataset->dims;

	//draw grid
    canvas->SetForeground(interface_color[GRID1]);
    canvas->SetXorMode(false);

    for (int i = 0;i < dataset->dims;i++)
        for (int j = 0;j < dataset->dims;j++)
            canvas->DrawRectangle(i*boxwidth, j*boxheight, boxwidth, boxheight);

	// draw labels
    canvas->SetForeground(interface_color[TEXT]);

    Vec2 font_size = canvas->GetDataFontSize("X");
    float max_str_size = boxwidth - font_size.X;

    // draw horizontal labels
    // add letters one at a time until space runs out
    for (i=0; i<dataset->dims; i++)
    {
        float x = i*boxwidth;
        float y = .8*font_size.Y;
        canvas->DrawConfinedString(x,y,dataset->names[i],max_str_size);
    }

    // draw vertical labels
    // add letters one at a time until space runs out
    for (i=0; i<dataset->dims; i++)
    {
        // initial coordinates
        Vec2 p;
        p.X = -1.5*font_size.X;
        p.Y = i*boxheight; // - font_size.Y*1.1;

        for (int j=0; j<int(strlen(dataset->names[i])); j++)
        {
            Vec2 letter = canvas->GetDataFontSize(&dataset->names[i][j],1);
            p.Y -= 1.2*letter.Y;

            if (p.Y > (i+1)*boxheight+letter.Y) break;

            canvas->DrawString(p.X,p.Y,&dataset->names[i][j],1);
        }
    }
}


//draw the band of a cluster
static void hierScatterDrawSingleNodeBand( grCanvas *canvas, Cluster &c, float density)
{
	bool is_brushed = BITTEST(c.flags,IS_BRUSHED);
    if (MAGNIFY && !is_brushed) return;
    if ( c.num_children == 0 ) return;
	//set color
    if (COLORED_BANDS) {

        int color_idx = (int)(c.color*(CMAPSIZE-1));

        if (is_brushed)  {
            DB(fade_factor);
            if (fade_factor>0)
                canvas->SetForeground( CMAP[ color_idx ], 1.0);
            else
                canvas->SetForeground( CMAP[ color_idx ], 1+fade_factor);
        }
        else {
            if (fade_factor>0)
                canvas->SetForeground( CMAP[ color_idx ], 1-fade_factor);
            else
                canvas->SetForeground( CMAP[ color_idx ], 1.0);
        }
    }
    else
        canvas->SetForeground(interface_color[BRUSH1+2]);

	//draw band
    ARR<Vec2> p(5);
    for (int j=0; j<dataset->dims; j++) 
	{
        for (int k=0; k<dataset->dims; k++)
		{
/*			radiusj = c.sqrt_radius * (dataset->all_dim_max[j]-dataset->all_dim_min[j]);
			radiusk = c.sqrt_radius * (dataset->all_dim_max[k]-dataset->all_dim_min[k]);
			p[0].X = XMAP(j,c.mean[j]);
			p[0].Y = YMAP(k,c.mean[k]);
			p[1].X = XMAP(j,c.mean[j]-radiusj);
			p[1].Y = YMAP(k,c.mean[k]-radiusk);
			p[2].X = XMAP(j,c.mean[j]-radiusj);
			p[2].Y = YMAP(k,c.mean[k]+radiusk);
			p[3].X = XMAP(j,c.mean[j]+radiusj);
			p[3].Y = YMAP(k,c.mean[k]+radiusk);
			p[4].X = XMAP(j,c.mean[j]+radiusj);
			p[4].Y = YMAP(k,c.mean[k]-radiusk);
*/
			p[0].X = XMAP(j,c.mean[j]);
			p[0].Y = YMAP(k,c.mean[k]);
			p[1].X = XMAP(j,c.min[j]);
			p[1].Y = YMAP(k,c.min[k]);
			p[2].X = XMAP(j,c.min[j]);
			p[2].Y = YMAP(k,c.max[k]);
			p[3].X = XMAP(j,c.max[j]);
			p[3].Y = YMAP(k,c.max[k]);
			p[4].X = XMAP(j,c.max[j]);
			p[4].Y = YMAP(k,c.min[k]);
			if ( fabs(1 - extent_scaling_fac ) > 1e-4 )
			{
				for ( int i = 1; i <= 4; i++ )
				{
					p[i].Y = (1-extent_scaling_fac)*p[0].Y + (extent_scaling_fac*p[i].Y);
					p[i].X = (1-extent_scaling_fac)*p[0].X + (extent_scaling_fac*p[i].X);
				}
            }

			canvas->DrawTransparentSquare2(p, 0.0, density, 0);

/*
			band[3] = band[2] = p0;
			band[1] = p1;
			band[0] = p2;
            canvas->DrawTransparentBand2(band, 0.0, density, 0);
			band[0] = p4;
            canvas->DrawTransparentBand2(band, 0.0, density, 0);
			band[1] = p3;
            canvas->DrawTransparentBand2(band, 0.0, density, 0);
			band[0] = p2;
            canvas->DrawTransparentBand2(band, 0.0, density, 0);
*/
		}
    }

}


// draw band of all clusters
static void hierScatterDrawColoredBands(grCanvas *canvas, int on_flags=0, int off_flags=0)
{
	//draw cluster's band one by one
	for (int i=0;i<btree->nodes.Len();i++)
    {
        Cluster &c = btree->nodes[i];
        if (!SCAN_FLAGS(c.flags, on_flags, off_flags)) continue;

        hierScatterDrawSingleNodeBand(canvas, c, 0.5);
    }
}

/*
static void hierScatterDrawSingleNodeMeanAndSolidBand( grCanvas *canvas, Cluster &c, int style)
{
// jing setting
    if (BITTEST(style,STYLE_NONE)) return;

    float opacity = 1.0;

//  if (BITTEST(style,STYLE_DIM)) opacity *= 0.75;

    if (BITTEST(style,STYLE_HIGHLIGHT)) {

        // Want to fade out the brushed lines
        if (fade_factor<0) opacity = 1+fade_factor;

        canvas->SetForeground(interface_color[SBB], opacity);
    } else
    if (BITTEST(style,STYLE_COLORMAP)) {

        // Want to fade out the unbrushed lines
        if (fade_factor>0) opacity=1-fade_factor;

        // Check if mean is covered by data brush.
        if ( c.mark == okcdata::COLORED )
            canvas->SetForeground(interface_color[HIGHLIGHT], opacity);
        else {
            int color_idx = int(c.color*(CMAPSIZE-1));
            canvas->SetForeground( CMAP[color_idx], opacity);
        }
    }
//jing setting

	canvas->BeginDraw();
	glBegin(GL_QUADS);

	double radiusj, radiusk;
    for (int j=0; j<dataset->dims; j++) 
	{
        for (int k=0; k<dataset->dims; k++)
		{
			radiusj = c.sqrt_radius * (dataset->all_dim_max[j]-dataset->all_dim_min[j]);
			radiusk = c.sqrt_radius * (dataset->all_dim_max[k]-dataset->all_dim_min[k]);
			glVertex2f( XMAP(j,c.mean[j]-radiusj), YMAP(k,c.mean[k]-radiusk) );
			glVertex2f( XMAP(j,c.mean[j]-radiusj), YMAP(k,c.mean[k]+radiusk) );
			glVertex2f( XMAP(j,c.mean[j]+radiusj), YMAP(k,c.mean[k]+radiusk) );
			glVertex2f( XMAP(j,c.mean[j]+radiusj), YMAP(k,c.mean[k]-radiusk) );
			
		}
    }

	glEnd();
	canvas->EndDraw();
}
*/

//draw the mean of a cluster
static void hierScatterDrawSingleNodeMean( grCanvas *canvas, Cluster &c, int style)
{
    if (BITTEST(style,STYLE_NONE)) return;

    //set color
	float opacity = 1.0;

//  if (BITTEST(style,STYLE_DIM)) opacity *= 0.75;

    if (BITTEST(style,STYLE_HIGHLIGHT)) {

        // Want to fade out the brushed lines
        if (fade_factor<0) opacity = 1+fade_factor;

        canvas->SetForeground(interface_color[SBB], opacity);
    } else
    if (BITTEST(style,STYLE_COLORMAP)) {

        // Want to fade out the unbrushed lines
        if (fade_factor>0) opacity=1-fade_factor;

        // Check if mean is covered by data brush.
        if ( c.mark == okcdata::COLORED )
            canvas->SetForeground(interface_color[HIGHLIGHT], opacity);
        else {
            int color_idx = int(c.color*(CMAPSIZE-1));
            canvas->SetForeground( CMAP[color_idx], opacity);
        }
    }

	//draw the mean of a cluster
	canvas->BeginDraw();
	glBegin(GL_POINTS);

    for (int j=0; j<dataset->dims; j++) 
	{
        for (int k=0; k<dataset->dims; k++)
		{
			glVertex2f( XMAP(j,c.mean[j]), YMAP(k,c.mean[k]) );
		}
    }

	glEnd();
	canvas->EndDraw();

}

/*
static void hierScatterDrawColoredMeansAndSolidBands(grCanvas *canvas, int style, int on_flags=0, int off_flags=0 )
{
    for (int i=0;i<btree->nodes.Len();i++)
    {
        Cluster &c = btree->nodes[i];
        if (!SCAN_FLAGS(c.flags, on_flags, off_flags)) continue;
		if ( fabs(c.sqrt_radius) < 1e-9 )
		    hierScatterDrawSingleNodeMean(canvas, c, style);
		else
			hierScatterDrawSingleNodeMeanAndSolidBand(canvas, c, style);
    }
}
*/

//draw means for all clusters
static void hierScatterDrawColoredMeans(grCanvas *canvas, int style, int on_flags=0, int off_flags=0 )
{
    for (int i=0;i<btree->nodes.Len();i++)
    {
        Cluster &c = btree->nodes[i];
        if (!SCAN_FLAGS(c.flags, on_flags, off_flags)) continue;
	    hierScatterDrawSingleNodeMean(canvas, c, style);
    }
}


void do_cluster_scatter( grCanvas *canvas )
{
    if (!dataset) return;

    canvas->SetTopLeftOrigin();
    grCanvas::grResizeMainCanvas(canvas);

    canvas->SetBackground(interface_color[BG]);
    canvas->Clear();

    hierInitClusterTree();

    if (!btree) return;

//draw brushes    
	//if (SHOW_BRUSH_BKGRD)
//    DrawScatterBrushes(canvas);

//draw grid and labels of the plots
	DrawScatterGridAndLabel(canvas);

    canvas->SetXorMode(false);


    if (SHOW_BANDS) {
        if (!MAGNIFY) hierScatterDrawColoredBands(canvas, IS_CONTOUR0);
        hierScatterDrawColoredBands(canvas, IS_CONTOUR1 | IS_BRUSHED);
    }

  /*  if (SHOW_MEANS) {
	    glPointSize((DSIZE+1)*canvas->zoom.X);
        if (!MAGNIFY) hierScatterDrawColoredMeansAndSolidBands(canvas, normal_style, IS_CONTOUR0, IS_BRUSHED );
        hierScatterDrawColoredMeansAndSolidBands(canvas, highlight_style, IS_CONTOUR1 | IS_BRUSHED );
	    glPointSize(1.0);
    }
*/
    if (SHOW_MEANS) {
	    glPointSize((DSIZE+1)*canvas->zoom.X);
        if (!MAGNIFY) hierScatterDrawColoredMeans(canvas, normal_style, IS_CONTOUR0, IS_BRUSHED );
        hierScatterDrawColoredMeans(canvas, highlight_style, IS_CONTOUR1 | IS_BRUSHED );
	    glPointSize(1.0);
    }
	
	/*
    // Draw average operations
    for (int op=0; op<num_oper; op++)
    {
        // Check if this operation has OPER_AVERAGE set 
        if (!(brush_ops[op].oper & AVERAGE)) continue;

        ARR<float> average(dataset->dims);
        if (BrushOperCompAverage(op, average)==0) continue;

        glLineWidth(3.0);
        canvas->SetForeground(interface_color[TEXT]);
        DrawParcoordAverage(canvas,average);
    }

    DrawParcoordAxes(canvas);*/
}


void InputHierScatterMouseMotion(grCanvas *canvas, int x, int y)
{
    // Hack to disable multiselect in HPC
    if (canvas->shiftkey && canvas->button[0] == true) return;

    InputScatterMouseMotion(canvas,x,y);
}

void InputHierScatterButtonPress(grCanvas *canvas, int button, int x, int y)
{
    // Hack to disable multiselect in HPC
    if (canvas->shiftkey && button == 0) return;

    InputScatterButtonPress(canvas,button,x,y);
}

void InputHierScatterButtonRelease(grCanvas *canvas, int button, int x, int y)
{
    // Hack to disable multiselect in HPC
    if (canvas->shiftkey && button == 0) return;

    InputScatterButtonRelease(canvas,button,x,y);

    if (canvas->shiftkey) return;

    if (!btree) return;

//    EXTENT_BRUSH = true;
//    hierClearNodeFlags(IS_BRUSHED);
//    hierModifyExtentsByBrushParameters();
//    hierMarkBrushedClustersByExtents(&btree->nodes[0]);
    canvas->Refresh();
}

#include "X11/keysym.h"

void InputHierScatterKeyRelease(grCanvas *canvas, int key)
{
    if (key == XK_Shift_L || key == XK_Shift_R) return;

    appCanvasInputKeyRelease(canvas,key);

    if (key == XK_Shift_L || key == XK_Shift_R) {
        // Exit multiselect mode
        //
//        EXTENT_BRUSH = true;
//        hierClearNodeFlags(IS_BRUSHED);
//        hierModifyExtentsByBrushParameters();
//        hierMarkBrushedClustersByExtents(&btree->nodes[0]);
        canvas->Refresh();
    }
}
