////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//      Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //
//
// dimstack.C - function to generate dimensional stacking display for
// XmdvTool
//
// This function displays the data using dimensional stacking.  Each dimension
// is discretized into a user-specified number of uniformly sized buckets.
// This is called the cardinality of the dimension.  Two dimensions are
// selected as the "slowest" or "outermost" dimensions, and a virtual display
// is created based on the cardinality of the 2 dimensions.  At each virtual
// pixel, we create an image based on 2 more dimensions.  This recursive
// embedding continues until all dimensions are accounted for.  This function
// uses a fixed size rectangle to display each data point.  This forces
// scrollbars to pop up in the display area if the resulting size is too large.
// Data points may map to the same location in the descrete space.  Data is
// displayed using this projection algorithm, and grid marks of varying
// intensity show the transitions between the buckets.  Darker grids indicate
// transitions in slower dimensions.
//

#include "dimstack.H"
#include "ogl.H"
#include "XmdvTool.H"
#include "brush.H"
#include "brush_oper.H"
#include "multi_select.H"
#include <string.h>

static float x_blocksize, y_blocksize, x_size, y_size;

void RegisterDimstackCanvas(grCanvas *canvas)
{
    canvas->display_proc = do_dimstack;
    canvas->reshape_proc = grCanvas::grResizeMainCanvas;
    canvas->buttonpress_proc = InputDimstackButtonPress;
    canvas->motion_proc = InputDimstackMouseMotion;
    canvas->buttonrelease_proc = InputDimstackButtonRelease;
    canvas->keypress_proc = appCanvasInputKeyPress;
    canvas->keyrelease_proc = appCanvasInputKeyRelease;
}

//
// At each level, the number of pixels between successive values for the
// dimension involved varies.  It depends on the minimum block size and
// the cardinalities of the dimensions embedded within it.  This in turn
// depends on the orientation and speed of each dimension.
//
static void comp_pixel_size(ARR<float> &size)
{
    float odd=1, even=1;

    // Compute product of cardinalities in each orientation
    int i;
    for (i = 0;i < dataset->dims;i=i+2)
        odd *= dataset->cardinality[i];

    for (i = 1;i < dataset->dims;i=i+2)
        even *= dataset->cardinality[i];

    odd = 1.0/odd;
    even = 1.0/even;

    // Now compute size of a virtual pixel for each dimension
    for (i = 0;i < dataset->dims;i++)   {
        size[i] = (i&1)?even:odd;
        for (int j = i+2; j < dataset->dims; j=j+2)
            size[i] *= dataset->cardinality[j];
    }

    x_blocksize = odd;
    y_blocksize = even;
}

//
// Determine if this value (val) for this dimension (dim) is covered by
// the brush.
//
static bool is_within(int val, int dim, int brush)
{
    float fval = val;
    float range = dataset->dim_max[dim] - dataset->dim_min[dim];
    float b_pos = dataset->cardinality[dim] * (brushes[brush].pos[dim]-dataset->dim_min[dim]) / range;
    float b_size = .5 * dataset->cardinality[dim] * brushes[brush].size[dim] / range;
    float start = b_pos - b_size;
    float end = b_pos + b_size;
    if (end > fval && end <= (fval+1.0)) return true;
    if (start >= fval && start < (fval+1.0)) return true;
    if (start < fval && end > (fval+1.0)) return true;
    return false;
}

//
// Recursive check for brush coverage, 2 dimensions at a time
//
static void rec_fill(grCanvas *canvas, float basex, float basey, int nextx, int nexty,
          ARR<float> &size, int brush)
{
    // If we had an even number of dimensions and are at the bottom of the
    // recursion, draw a rectangle in that data spot, as it is covered.
    //
    if (nextx == dataset->dims)
    {
        canvas->FillRectangle(basex, y_size - basey - y_blocksize, x_blocksize, y_blocksize);
        return;
    }

    // If we had an odd number of dimensions and are at the bottom of the
    // recursion, check each descrete value for brush overlap and draw rectangles
    // over covered spots.
    //
    int i;
    if (nexty == dataset->dims)
    {
        for (i=0; i<dataset->cardinality[nextx]; i++) {
            if (is_within(i, nextx, brush) == true)
                canvas->FillRectangle(basex + i * size[nextx],
                    y_size - basey - y_blocksize, x_blocksize, y_blocksize);
        }
        return;
    }

    // Continue recursion for each bucket of the current 2 dimensions
    //
    for (i=0; i<dataset->cardinality[nextx]; i++)
        for (int j=0; j<dataset->cardinality[nexty]; j++)
            if ((is_within(i, nextx, brush) == 1) && (is_within(j, nexty, brush) == 1))
                rec_fill(canvas, basex + i * size[nextx], basey + j * size[nexty],
                    nextx+2, nexty+2, size, brush);
}

//
// Display brush coverage by descretizing brush in each dimension and check
// where in range of that dimension it overlaps.  Roundoff error in the
// discretization process can reduce the visual effect of brush coverage.
//
void ds_outline(grCanvas *canvas, ARR<float> &size)
{
    for (int brush=0; brush<MAXBRUSH; brush++)
    {
        // If this brush is not being displayed, skip it
        if (brushes[brush].display == false) continue;

        // Set the appropriate brush color
        canvas->SetForeground(interface_color[BRUSH1+brush]);

        // For the outermost dimensions...
        for (int i=0; i<dataset->cardinality[0]; i++) {
            for (int j=0; j<dataset->cardinality[1]; j++)
            {
                // if the brush intersects those dimensions recursively
                // call to fill
                if ((is_within(i, 0, brush) == 1) && (is_within(j, 1, brush) == 1))
                    rec_fill(canvas, i * size[0], j * size[1], 2, 3, size, brush);
            }
        }
    }
}

//
// Compute the descretized value from an original data value for a particular
// dimension.  Display will be messed up if it is out of bounds.  You then
// should fix the range for that dimension in your data file.
//
int get_contrib(int ind, float data)
{
    int res = int( ((data - dataset->dim_min[ind]) / (dataset->dim_max[ind] - dataset->dim_min[ind])) *
              (float)(dataset->cardinality[ind]));
/*
    if (res >= dataset->cardinality[ind] || res < 0) {
        printf("Data point out of bounds: dimension %d: [%f, %f]: %f\n", ind,
            dataset->dim_min[ind], dataset->dim_max[ind], data);
    }
*/
    return res;
}

void DrawSingleDimstackBin(grCanvas *canvas, const ARR<float> &data, const ARR<float> &size)
{
    ARR<int> contrib(dataset->dims);
    int j;
    for (j=0; j<dataset->dims; j++)
        contrib[j] = get_contrib(j, data[j]);

    float endx = 0, endy = 0;
    for (j=0; j<dataset->dims; j++)
    {
        endx += contrib[j] * size[j];
        j++;
        if (j < dataset->dims)
            endy += contrib[j] * size[j];
    }
    canvas->FillRectangle(endx, y_size - endy - y_blocksize, x_blocksize, y_blocksize);
}


//
// Draw an average value on the dimensional stacking
// display.
//
static void DrawDimstackAverage(grCanvas *canvas, const ARR<float> &data)
{
    // Compute the size of the blocks needed for the slowest dimension
    ARR<float> size(dataset->dims);
    comp_pixel_size(size);

    x_size = size[0]*dataset->cardinality[0];
    y_size = size[1]*dataset->cardinality[1];

    // Each dimension will contribute to the position calculation
    ARR<float> contrib(dataset->dims);
    int j;
    for (j = 0;j < dataset->dims;j++)
        contrib[j] = get_contrib(j, data[j]);

    float endx = 0, endy = 0;
    for (j = 0;j < dataset->dims;j++)
    {
        endx += contrib[j] * size[j];
        j++;
        if (j < dataset->dims)
        endy += contrib[j] * size[j];
    }

    canvas->FillRectangle(endx, y_size - endy - y_blocksize, x_blocksize, y_blocksize);
}

void do_dimstack(grCanvas *canvas)
{
    if (!dataset) return;

    if (dataset->dims<2) {
        fprintf(stderr,"Warning: Dimension stacking needs at least 2 active dimensions.\n");
        return;
    }

    canvas->SetTopLeftOrigin();
    grCanvas::grResizeMainCanvas(canvas);

    canvas->SetBackground(interface_color[BG]);
    canvas->Clear();

    // Compute the size of the blocks needed for the slowest dimension
    //
    ARR<float> size(dataset->dims);
    comp_pixel_size(size);
    x_size = size[0]*dataset->cardinality[0];
    y_size = size[1]*dataset->cardinality[1];

    // Display the brush coverage if desired
    ds_outline(canvas, size);

    // Draw up to 3 levels of grid marks, based on the number of dimensions
    if (dataset->dims >= 6)
    {
        canvas->SetForeground(interface_color[GRID3]);
        for (float x = 0;x <= x_size; x = x + size[4])
            canvas->DrawLine(x, 0, x, y_size);

        for (float y = 0;y <= y_size; y = y + size[5])
            canvas->DrawLine(0, y, x_size, y);
    }

    if (dataset->dims >= 4)
    {
        canvas->SetForeground(interface_color[GRID2]);
        for (float x = 0;x <= x_size; x = x + size[2])
            canvas->DrawLine(x, 0, x, y_size);

        for (float y = 0;y <= y_size; y = y + size[3])
            canvas->DrawLine(0, y, x_size, y);
    }

    canvas->SetForeground(interface_color[GRID1]);

    for (float x = 0;x <= x_size; x = x + size[0])
        canvas->DrawLine(x, 0, x, y_size);

    for (float y = 0;y <= y_size; y = y + size[1])
        canvas->DrawLine(0, y, x_size, y);

    // Now do the same thing with the data points being touched by the brush

    // First draw all points in the data color
    //
    ARR<float> data(dataset->dims);
    ARR<int> contrib(dataset->dims);

    canvas->SetForeground(interface_color[DATA]);

    int i;
    for (i=0; i<dataset->data_size; i++)
    {
        // First time through only paint regular data
        //
        if (dataset->mark[i] != okcdata::REGULAR) continue;

        dataset->GetData(data, i);
        DrawSingleDimstackBin(canvas,data,size);
    }

    // Now paint all points covered by an operation
    //
    for (i=0; i<dataset->data_size; i++)
    {
        if (dataset->mark[i] == okcdata::DISABLED) continue;
        if (dataset->mark[i] == okcdata::REGULAR) continue;

        dataset->GetData(data, i);

        canvas->SetForeground(dataset->color[i]);
        DrawSingleDimstackBin(canvas,data,size);
    }

    // Draw average operations
    for (int op=0; op<num_oper; op++)
    {
        // Check if this operation has OPER_AVERAGE set */
        if (!(brush_ops[op].oper & AVERAGE)) continue;

        ARR<float> average(dataset->dims);
        if (BrushOperCompAverage(op, average)==0) continue;

//      glLineWidth(3.0);
        canvas->SetForeground(interface_color[TEXT]);
        DrawDimstackAverage(canvas,average);
    }
//    glPointSize(1.0);
}

static void DimstackMapScreenToData(float x,float y, ARR<float> &center)
{
    // Determine size of each dimension
    ARR<float> size(dataset->dims);
    comp_pixel_size(size);

    // Now find the bin that the user clicked in
    int i;
    for (i=0; i<dataset->dims; i++)
    {
        // If odd, it's in the y direction
        if (i%2)
        {
            int bin = int(y/size[i]);

            // invert y
            int orig_bin = bin;
            bin = dataset->cardinality[i] - 1 - bin;

            // check bounds
            if (bin < 0 || bin >= dataset->cardinality[i])
                return;

            // Increment data coords based on this
            center[i] = bin * (dataset->dim_max[i] - dataset->dim_min[i])/dataset->cardinality[i];
            center[i] += dataset->dim_min[i];

            // use original bin # for y decrement
            y -= orig_bin*size[i];
        }

        // If even, it's in the x direction
        else
        {
            int bin = int(x/size[i]);

            // check bounds
            if (bin < 0 || bin >= dataset->cardinality[i])
                return;

            // Increment data coords based on this
            center[i] = bin * (dataset->dim_max[i] - dataset->dim_min[i])/dataset->cardinality[i];
            center[i] += dataset->dim_min[i];

            x -= bin*size[i];
        }
    }
}

static void InputDimstackBrushSelect(grCanvas *canvas, int sx, int sy)
{
    Vec2 p = canvas->MapScreenToData( Vec2(sx,sy) );

    // If brushing is not on or not displayed, forget it
    if (curr_brush->display == false || curr_brush->enable == false)
        return;

    // Now find the bin that the user clicked in
    ARR<float> new_center(dataset->dims);
    DimstackMapScreenToData(p.X,p.Y,new_center);

    // Recenter the brush
    int i;
    for (i=0; i<dataset->dims; i++)
        curr_brush->pos[i] = new_center[i];

    // Check if any of the brush values went out of range
    for (i=0; i<dataset->dims; i++)
    {
        // don't resize brush, but just move it so it stays in range
        if (curr_brush->Max(i) > dataset->dim_max[i])
            curr_brush->pos[i] = dataset->dim_max[i] - curr_brush->size[i]/2.;
        if (curr_brush->Min(i) < dataset->dim_min[i])
            curr_brush->pos[i] = dataset->dim_min[i] + curr_brush->size[i]/2.;
    }
}

static void InputDimstackMultiSelect(grCanvas *canvas, int sx, int sy)
{
    if (!multi_select) return;

    Vec2 p = canvas->MapScreenToData( Vec2(sx,sy) );

    // Now find the bin that the user clicked in
    ARR<float> center(dataset->dims);
    DimstackMapScreenToData(p.X,p.Y,center);

/*
    Vec2 screen_rad = canvas->MapScreenToData( Vec2(multisel_radius,multisel_radius) ) -
                      canvas->MapScreenToData( Vec2(0,0) );
*/
    const float radN = .8;

    ARR<float> data(dataset->dims);

    // Find all points within circle distance of mouse
    //
    for (int i=0; i<dataset->data_size; i++)
    {
        dataset->GetData(data, i);

        float dist = 0;
        for (int j=0;j<dataset->dims;j++) {
            float d = (center[j] - data[j]);
            dist += d*d;
        }
        dist = sqrt(dist);

        if (dist < radN)
        {
            multiSelectAddPoint(i);

            ARR<float> size(dataset->dims);
            comp_pixel_size(size);
            glDrawBuffer(GL_FRONT);
            canvas->SetForeground(interface_color[PAINT]);
            DrawSingleDimstackBin(canvas,data,size);
            glDrawBuffer(GL_BACK);
        }
    }
}

static void DimstackCaption(grCanvas *canvas, int x, int y)
{
    Vec2 p = canvas->MapScreenToData( Vec2(x,y) );

    // Now find the bin that the user clicked in
    ARR<float> center(dataset->dims);
    DimstackMapScreenToData(p.X,p.Y,center);

    char caption[500]="Bin center: (";

    for (int i=0; i<dataset->dims; i++) {
        char str[100];
        sprintf(str, "%s: %.2f ", dataset->names[i], center[i]);
        strcat(caption, str);
    }
    strcat(caption,")");

    appShowCaption(Togl_Interp(canvas->GetWidget()),caption);
}

void InputDimstackButtonPress(grCanvas *canvas, int button, int x, int y)
{
    if (!dataset) return;

    if (canvas->shiftkey && button == 0) {
        InputDimstackMultiSelect(canvas,x,y);
        return;
    } else

    if (button==2) {
        canvas->SetHandCursor();
        return;
    } else

    if (button == 0) {
        InputDimstackBrushSelect(canvas,x,y);
        BrushOperMarkDataset();
        appRefreshAllCanvas(Togl_Interp(canvas->GetWidget()));
    }
}

void InputDimstackMouseMotion(grCanvas *canvas, int x, int y)
{
    if (!dataset) return;

    DimstackCaption(canvas, x, y);

    if (canvas->shiftkey && canvas->button[0] == true) {
        InputDimstackMultiSelect(canvas,x,y);
        return;
    } else

    if (canvas->button[2] == true) {
        appCanvasMousePan(canvas);
    }
}

void InputDimstackButtonRelease(grCanvas *canvas, int button, int x, int y)
{
    if (canvas->shiftkey && button == 0) {
        return;
    }
    canvas->SetDefaultCursor();
}
