////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//      Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //
//
// glyph.C - function for generating glyphs for XmdvTool
//
// Draw data using a star glyph pattern.  Each dimension of a data point
// controls the length of a ray emanating from a point, and the ends of all
// N rays are connected with a polyline.  Glyphs are currently spaced
// evenly across the drawing area, using the square root of the data set
// size to determine the number of rows and columns of glyphs.  We could also
// use 2 dimensions to control the location of the glyphs, but you tend to
// get a lot of overlap.

#include "glyph.H"
#include "ogl.H"
#include "XmdvTool.H"
#include "brush.H"
#include "brush_oper.H"
#include "multi_select.H"
#include <string.h>

void RegisterGlyphCanvas(grCanvas *canvas)
{
    canvas->display_proc = do_glyph;
    canvas->reshape_proc = grCanvas::grResizeMainCanvas;
    canvas->buttonpress_proc = InputGlyphButtonPress;
    canvas->motion_proc = InputGlyphMouseMotion;
    canvas->buttonrelease_proc = InputGlyphButtonRelease;
    canvas->keypress_proc = appCanvasInputKeyPress;
    canvas->keyrelease_proc = appCanvasInputKeyRelease;
}

void DrawGlyph(grCanvas *canvas, const ARR<float> &data, float basex, float basey, float blocksize)
{
    ARR<Vec2> poly(dataset->dims+1);
    ARR<float> norm(dataset->dims);

    // Compute normalized length of each ray for sample
    //
    int j;
    for (j=0; j<dataset->dims; j++) {
        norm[j] = (data[j] - dataset->dim_min[j]) * (blocksize/2.0) /
                (dataset->dim_max[j] - dataset->dim_min[j]);
    }

    // Compute endpoints for each ray based on evenly spaced angles
    // from center
    //
    float alpha = 2.0*M_PI/dataset->dims;

    for (j=0; j<dataset->dims; j++)
    {
        poly[j].X = basex + norm[j] * cos(j*alpha);
        poly[j].Y = basey - norm[j] * sin(j*alpha);
        canvas->DrawLine(basex, basey, poly[j].X, poly[j].Y);
    }

    // Draw bounding polyline
    //
    poly[dataset->dims].X = poly[0].X;
    poly[dataset->dims].Y = poly[0].Y;

    canvas->DrawLines(poly);
}

//
// Redraws a single glyph on the canvas.  Usefull when
// doing multi-select for example.
//
static void RedrawSingleGlyph(grCanvas *canvas, int glyphno)
{
    ARR<float> data(dataset->dims);
    dataset->GetData(data, glyphno);

    DrawGlyph(canvas,data,dataset->basex[glyphno],dataset->basey[glyphno],dataset->blocksize);
}

//
// Draw an average value on the glyph display.
//
static void DrawGlyphAverage(grCanvas *canvas, int glyphno, const ARR<float> &data)
{
    // Get location for glyph
//    float basex = blocksize * (glyphno%blockcnt) + blocksize/2;
//    float basey = blocksize * (glyphno/blockcnt) + blocksize/2;

    DrawGlyph(canvas,data,dataset->basex[glyphno],dataset->basey[glyphno],dataset->blocksize);
}

void do_glyph(grCanvas *canvas)
{
    if (!dataset) return;

    canvas->SetTopLeftOrigin();
    grCanvas::grResizeMainCanvas(canvas);

    canvas->SetBackground(interface_color[BG]);
    canvas->Clear();

    // No text or brush coverage drawn, so just choose colors based on whether
    // brushed points will be displayed or not
    //
    ARR<float> data(dataset->dims);

    glLineWidth(DSIZE);

    for (int i=0; i<dataset->data_size; i++)
    {
        if (dataset->mark[i] == okcdata::DISABLED) continue;

        if (dataset->mark[i] == okcdata::REGULAR) {
            canvas->SetForeground(interface_color[DATA]);
        } else {
            canvas->SetForeground(dataset->color[i]);
        }

//        float basex = blocksize * (i%blockcnt) + halfblocksize;
        //float basey = blocksize * (i/blockcnt) + halfblocksize;

        dataset->GetData(data, i);
        DrawGlyph(canvas,data,dataset->basex[i],dataset->basey[i],dataset->blocksize);
    }
    glLineWidth(1.0);

    // Draw average operations
    for (int op=0; op<num_oper; op++)
    {
        // Check if this operation has OPER_AVERAGE set */
        if (!(brush_ops[op].oper & AVERAGE)) continue;

        ARR<float> average(dataset->dims);
        if (BrushOperCompAverage(op, average)==0) continue;

        glLineWidth(2.0);
        canvas->SetForeground(interface_color[TEXT]);
        DrawGlyphAverage(canvas,dataset->data_size+op,average);
    }
    glLineWidth(1.0);
}

/*
static void InputGlyphBrushSelect(grCanvas *canvas, int sx, int sy)
{
    Vec2 p = canvas->MapScreenToData( Vec2(sx,sy) );

    // If brushing is not on or not displayed, forget it
    if (curr_brush->display == false || curr_brush->enable == false)
        return;

    // Figure out how big each glyph will be by dividing up the
    // drawing area
    float blocksize = 1.0/(1.0 + sqrt(dataset->data_size));
    int blockcnt = int( .5 + sqrt(dataset->data_size) );

    // Check if in range
    if (p.Y / blocksize >= blockcnt || p.X / blocksize >= blockcnt)
      return;

    // Map mouse position to glyph
    int glyph_no = int(p.Y / blocksize);
    glyph_no *= blockcnt;
    glyph_no += int(p.X / blocksize);

    if (glyph_no < 0 || glyph_no >= dataset->data_size) return;

    if (multi_select)
    {
        // Delete point if already there
        if (multiSelectDeletePoint(glyph_no)==true)
        {
            glDrawBuffer(GL_FRONT);
            canvas->SetForeground(interface_color[DATA]);
            RedrawSingleGlyph(canvas, glyph_no);
            glDrawBuffer(GL_BACK);
        } else
        {
            // add this point
            multiSelectAddPoint(glyph_no);
            glDrawBuffer(GL_FRONT);
            canvas->SetForeground(interface_color[PAINT]);
            RedrawSingleGlyph(canvas, glyph_no);
            glDrawBuffer(GL_BACK);
        }
        return;
    }

    ARR<float> data(dataset->dims);
    dataset->GetData(data, glyph_no);

    int i;
    for (i=0; i<dataset->dims; i++)
        curr_brush->pos[i] = data[i];

    // Check if any of the brush values went out of range
    for (i=0; i<dataset->dims; i++)
    {
        // don't resize brush, but just move it so it stays in range
        if (curr_brush->Max(i) > dataset->dim_max[i])
            curr_brush->pos[i] = dataset->dim_max[i] - curr_brush->size[i]/2.;
        if (curr_brush->Min(i) < dataset->dim_min[i])
            curr_brush->pos[i] = dataset->dim_min[i] + curr_brush->size[i]/2.;
    }
}
*/

static void InputGlyphBrushSelect(grCanvas *canvas, int sx, int sy)
{
    Vec2 p = canvas->MapScreenToData( Vec2(sx,sy) );

    // If brushing is not on or not displayed, forget it
    if (curr_brush->display == false || curr_brush->enable == false)
        return;

    // Figure out how big each glyph will be by dividing up the
    // drawing area
    float halfbsize = 0.5/(1.0 + sqrt(dataset->data_size));

	int i, glyph_no = -1;
	float min = halfbsize *3, temp;
	for ( i = 0; i < dataset->data_size; i++ )
	{
		if ( (fabs(p.X - dataset->basex[i]) < halfbsize) && 
			 (fabs(p.Y - dataset->basey[i]) < halfbsize) ) 
		{
			temp = fabs(p.X - dataset->basex[i]) + fabs(p.Y - dataset->basey[i]);
			if ( temp < min )
			{
				glyph_no = i; 
				min = temp;
			}
		}
	}

	if ( glyph_no == -1 )
		return;

    if (multi_select)
    {
        // Delete point if already there
        if (multiSelectDeletePoint(glyph_no)==true)
        {
            glDrawBuffer(GL_FRONT);
            canvas->SetForeground(interface_color[DATA]);
            RedrawSingleGlyph(canvas, glyph_no);
            glDrawBuffer(GL_BACK);
        } else
        {
            // add this point
            multiSelectAddPoint(glyph_no);
            glDrawBuffer(GL_FRONT);
            canvas->SetForeground(interface_color[PAINT]);
            RedrawSingleGlyph(canvas, glyph_no);
            glDrawBuffer(GL_BACK);
        }
        return;
    }

    ARR<float> data(dataset->dims);
    dataset->GetData(data, glyph_no);

    for (i=0; i<dataset->dims; i++)
        curr_brush->pos[i] = data[i];

    // Check if any of the brush values went out of range
    for (i=0; i<dataset->dims; i++)
    {
        // don't resize brush, but just move it so it stays in range
        if (curr_brush->Max(i) > dataset->dim_max[i])
            curr_brush->pos[i] = dataset->dim_max[i] - curr_brush->size[i]/2.;
        if (curr_brush->Min(i) < dataset->dim_min[i])
            curr_brush->pos[i] = dataset->dim_min[i] + curr_brush->size[i]/2.;
    }
}

void InputGlyphButtonPress(grCanvas *canvas, int button, int x, int y)
{
    if (!dataset) return;

    if (button==2) {
        canvas->SetHandCursor();
        return;
    }

    if (button == 0) {
        InputGlyphBrushSelect(canvas,x,y);
        BrushOperMarkDataset();
    }

    if (!multi_select)
        appRefreshAllCanvas(Togl_Interp(canvas->GetWidget()));
}

/*
static void GlyphCaption(grCanvas *canvas, int x, int y)
{
    Vec2 p = canvas->MapScreenToData( Vec2(x,y) );

    // Figure out how big each glyph will be by dividing up the
    // drawing area
    float blocksize = 1.0/(1.0 + sqrt(dataset->data_size));
    int blockcnt = int( .5 + sqrt(dataset->data_size) );

    // Check if in range
    if (p.Y / blocksize >= blockcnt || p.X / blocksize >= blockcnt)
      return;

    // Map mouse position to glyph
    int glyph_no = int(p.Y / blocksize);
    glyph_no *= blockcnt;
    glyph_no += int(p.X / blocksize);

    if (glyph_no < 0 || glyph_no >= dataset->data_size) return;

    ARR<float> data(dataset->dims);
    dataset->GetData(data, glyph_no);

    char caption[500]="\0";
    sprintf(caption, "Glyph %d: (", glyph_no);

    for (int i=0; i<dataset->dims; i++) {
        char str[100];

        if (dataset->nominal_fields[i] != 0) {
//            int value = (int) rint(data[i]);
//jing
            int value = (int)(data[i]+0.5);
//jing
            const char *string = dataset->nominal_fields[i]->MapValueToString(value);
            sprintf(str, "%s=%s ", dataset->names[i], string);
        } else {
            sprintf(str, "%s=%.2f ", dataset->names[i], data[i]);
        }

        strcat(caption, str);
    }
    strcat(caption,")");

    appShowCaption(Togl_Interp(canvas->GetWidget()),caption);
}
*/


static void GlyphCaption(grCanvas *canvas, int x, int y)
{
    Vec2 p = canvas->MapScreenToData( Vec2(x,y) );

    // Figure out how big each glyph will be by dividing up the
    // drawing area
    float halfbsize = 0.5/(1.0 + sqrt(dataset->data_size));

	int i, glyph_no = -1;
	float min = halfbsize *3, temp;
	for ( i = 0; i < dataset->data_size; i++ )
	{
		if ( (fabs(p.X - dataset->basex[i]) < halfbsize) && 
			 (fabs(p.Y - dataset->basey[i]) < halfbsize) ) 
		{
			temp = fabs(p.X - dataset->basex[i]) + fabs(p.Y - dataset->basey[i]);
			if ( temp < min )
			{
				glyph_no = i; 
				min = temp;
			}
		}
	}

	if ( glyph_no == -1 )
		return;

    ARR<float> data(dataset->dims);
    dataset->GetData(data, glyph_no);

    char caption[500]="\0";
    sprintf(caption, "Glyph %d: (", glyph_no);

    for (i=0; i<dataset->dims; i++) {
        char str[100];

        if (dataset->nominal_fields[i] != 0) {
//            int value = (int) rint(data[i]);
//jing
            int value = (int)(data[i]+0.5);
//jing
            const char *string = dataset->nominal_fields[i]->MapValueToString(value);
            sprintf(str, "%s=%s ", dataset->names[i], string);
        } else {
            sprintf(str, "%s=%.2f ", dataset->names[i], data[i]);
        }

        strcat(caption, str);
    }
    strcat(caption,")");

    appShowCaption(Togl_Interp(canvas->GetWidget()),caption);
}



void InputGlyphMouseMotion(grCanvas *canvas, int x, int y)
{
    if (!dataset) return;

    GlyphCaption(canvas, x, y);

    if (canvas->button[2] == true) {
        appCanvasMousePan(canvas);
    }
}

void InputGlyphButtonRelease(grCanvas *canvas, int button, int x, int y)
{
    if (canvas->shiftkey) {
        return;
    }

    canvas->SetDefaultCursor();
}


