////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//      Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //

#include "okcdata.H"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "macros.H"
#include "brush.H"
#include <math.h>
#include <search.h>
#include "XmdvTool.H"


//used for glyph placement

struct SortStruct{
	int number;
	float value;
};

static int IncreaseSort( const void *elm1, const void *elm2)
{
	float value1 = ((SortStruct *)elm1)->value;
	float value2 = ((SortStruct *)elm2)->value;

	if ( value1 < value2 )
		return -1;
	else if ( value1 == value2 )
		return 0;
	else
		return 1;
}


static int DecreaseSort( const void *elm1, const void *elm2)
{
	float value1 = ((SortStruct *)elm1)->value;
	float value2 = ((SortStruct *)elm2)->value;

	if ( value1 < value2 )
		return 1;
	else if ( value1 == value2 )
		return 0;
	else
		return -1;
}


//decide the base of every glyph according to their order in one dimension
void okcdata::Placement_Sort_Glyph(int dimNumber, bool increase)
{
	SortStruct *sortArray = new SortStruct[data_size];
	
	int i, k;
	k = dimNumber;
	for ( i = 0; i < data_size; i++ )			
	{
		sortArray[i].number = i;
		sortArray[i].value = data_buf[k];
		k = k+all_dims;
	}

	if ( increase == true )
		qsort( sortArray, data_size, sizeof( SortStruct ), IncreaseSort );
	else 
		qsort( sortArray, data_size, sizeof( SortStruct ), DecreaseSort );

    blocksize = 1.0/(1.0 + sqrt(data_size));
    blockcnt = int( .5 + sqrt(data_size) );
    halfblocksize = .5 * blocksize;

	for ( i = 0; i < data_size; i++ )
	{
		basex[sortArray[i].number] = blocksize * (i%blockcnt) + halfblocksize;
		basey[sortArray[i].number] = blocksize * (i/blockcnt) + halfblocksize;
	}

	delete sortArray;
	
}


//decide the base of every glyph according to their order in the file.
void okcdata::Placement_Original_Glyph()
{
    blocksize = 1.0/(1.0 + sqrt(data_size));
    blockcnt = int( .5 + sqrt(data_size) );
    halfblocksize = .5 * blocksize;

	for ( int i = 0; i < data_size; i++ )
	{
		basex[i] = blocksize * (i%blockcnt) + halfblocksize;
		basey[i] = blocksize * (i/blockcnt) + halfblocksize;
	}
	
}


//decide the base of every glyph according to their order in the file.
void okcdata::Placement_DataDriven_Glyph(int XdimNumber, int YdimNumber)
{
	float xmin = all_dim_min[XdimNumber]; 
	float ymin = all_dim_min[YdimNumber]; 
	float xpara = 1.0/(all_dim_max[XdimNumber] - xmin);
	float ypara = 1.0/(all_dim_max[YdimNumber] - ymin);
	
	int base = 0, dim = all_dims;
	for ( int i = 0; i < data_size; i++ )
	{
		basex[i] = (data_buf[base+XdimNumber] - xmin) * xpara;
		basey[i] = 1-(data_buf[base+YdimNumber] - ymin) * ypara;
		base += all_dims;
	}
	
}

/*************************************************************************/
//Code for glyph shift

/* program to adjust the positions of glyphs to reduce or eliminate overlaps.
Overlapping points exert a force at each other in the direction of the
vector between them, weighted by how close they are to each other.  Multiple
overlaps are simply summed and averaged.  Written by Matt Ward, last changed 8/2/00
*/

#define MAX_ITER 10             /* maximum number of iterations of adjustment(change as needed) */
#define SIZE 25                 /* maximum size of a glyph in x or y (change as needed) */
#define TOLERANCE 2             /* maximum allowable percentage of glyph overlaps (can be changed) */
#define SPEED SIZE/4            /* maximum change in position per iteration (arbitrary)*/

/* macro to determine if both x and y extents overlap between 2 glyphs - may have it backwards */       
#define OVER(a, b)  (((abs(pos[a].x-pos[b].x) < SIZE) && (abs(pos[a].y-pos[b].y) < SIZE))? 1:0)

struct position {               /* structure to hold a position */
   int x, y;
   };

struct position *pos;           /* buffer to hold glyph positions */
struct position *tpos;          /* temporary buffer for positions */

int num_pts;                    /* number of data points */

/* initialize space for positions and store them */
void init(int pts, float *basex, float *basey, double para)
{
	num_pts = pts;
	pos = (struct position *)malloc(pts * sizeof(struct position));
	tpos = (struct position *)malloc(pts * sizeof(struct position));
	if(pos == NULL || tpos == NULL)
	{
	  printf("Malloc failed in init\n");
	  exit(-1);
	}
/* now fill in positions of glyphs, whether data driven or derived */
/* CODE TO INITIALIZE POSITIONS GOES HERE */
	for ( int i = 0; i < pts; i++ )
	{
		pos[i].x = tpos[i].x = int ( 0.5 + basex[i]*para);
		pos[i].y = tpos[i].y = int ( 0.5 + basey[i]*para);
	}
}

/* count number of overlaps based on overlapping extents */
int num_overlaps(void)
{
   int i, j, num = 0;
   for(i = 0;i < num_pts;i++)
      for(j = i+1;j < num_pts;j++)
         num += OVER(i, j);
   return(num);
}

/* need to determine which points to move, what direction, and how much.
Points to move are those that overlap.  The direction of motion for point i is the
combined forces from all points that overlap it, formed as the vectors going
from the overlapping point to point i.  The vectors are weighted, so closer
points have more influence on the direction.  The overall motion is weighted by
SPEED, the maximum number of pixels a glyph can move */
void move_glyphs(void)
{
	int i, j, num;
	double mag;
	float vector[2], sumvector[2], maxsize;
	maxsize = (float)SIZE * 1.415;       /* max distance between overlap centers */
	sumvector[0] = sumvector[1] = 0.0;
	for(i = 0;i < num_pts;i++)   
	{
		num = 0;
/* for every point... */
		for(j = 0;j < num_pts;j++)        
		{
/* if it has any overlap */
			if(i != j && OVER(i, j))       
			{
				num++;
/* calculate vector between centers (the direction of force repelling points) */

/* if points are at same location, add a random displacement */
				if((pos[j].x == pos[i].x) && (pos[j].y == pos[i].y)) 
				{
					vector[0] = -1.0 + 2.0 * rand();        /* rand num between -1 and 1 */
					vector[1] = -1.0 + 2.0 * rand();
					mag = sqrt((vector[0] * vector[0]) + (vector[1] * vector[1]));
					/* normalize, and add with full weight to sumvector */
					vector[0] = vector[0]/mag;
					vector[1] = vector[1]/mag;
					sumvector[0] += vector[0];
					sumvector[1] += vector[1];
				}
/* normalize vector and add weighted amount to sumvector */
				else 
				{
					vector[0] = pos[i].x - pos[j].x;
					vector[1] = pos[i].y - pos[j].y;
					mag = sqrt((vector[0] * vector[0]) + (vector[1] * vector[1]));
					vector[0] = vector[0]/mag;
					vector[1] = vector[1]/mag;
					/* weight decreases as distance increases, from 1.0 to 0.0 */
					sumvector[0] += (1.0-mag/maxsize) * vector[0];
					sumvector[1] += (1.0-mag/maxsize) * vector[1];
				}
			}
		}
/* NOTE: may want to add a minor random offset for points being moved, to avoid
   oscillating between 2 locations */

/* average the vectors (this is not exactly correct, as lots of many nearby
points can dillute influence of very close one, resulting in smaller move, but in
correct direction)  */
		if(num > 0)       
		{
			sumvector[0] = sumvector[0]/(float)num;
			sumvector[1] = sumvector[1]/(float)num;
		}
/* calculate the new position for glyph i */
		tpos[i].x = pos[i].x + SPEED * sumvector[0];
		tpos[i].y = pos[i].y + SPEED * sumvector[1];
	}
/* now copy everything back into pos array */
	for(i = 0;i < num_pts;i++)   
	{
		pos[i].x = tpos[i].x;
		pos[i].y = tpos[i].y;
	}
}


/* compute tolerance as number of allowable overlaps and iteratively move
     points until tolerance met or maximum iterations reached */
void adjust(void)
{
   int i, tol;
   tol = TOLERANCE * num_pts / 100; 
   for(i = 0;i < MAX_ITER && num_overlaps() > tol;i++)
      move_glyphs();
}

void finished(int pts, float *basex, float *basey, double para)
{
//put back the glyph basex and basey
	for ( int i = 0; i < pts; i++ )
	{
		basex[i] = pos[i].x/para;
		basey[i] = pos[i].y/para;
	}
}


//decide the base of every glyph after reducing overlap.
void okcdata::Placement_Shift_Glyph()
{
	double para = SIZE/blocksize;
	init(data_size, basex, basey, para);	
	adjust();
	finished(data_size, basex, basey, para);
}


//////////////////////////////////////////////////////////////////////////////////
// code for space the glyphs

#define MAX_SPACE 4 /* maximum number of blank glyphs to put between glyphs */

static int *spaces, total_space;       /* spaces between adjacent points, and total space */
static float *dist;                    /* distances between adjacent points */

/* calculate the distance between 2 glyphs as the square root of the sum of the normalized
   squared difference in all the dimensions */

double comp_dist(ARR<float> dataA, ARR<float> dataB, ARR<float> maxmin, int numdims)
{
	double result = 0;
	int i;
	for(i = 0;i < numdims;i++)
	{
		if ( fabs(maxmin[i]) > 1e-9 )
			result += ((dataA[i] - dataB[i]) * (dataA[i] - dataB[i])) / maxmin[i];
	}
	return(sqrt(result));
}

void okcdata::Placement_Original_Space_Glyph()
{
  int i, index;
  int size, halfsize;   /* size of a glyph, and half the size, to center
position */
  float maxdist = 0.0;
  int row_length;       /* number of glyphs that will fit on a row */

/* get memory for lists */
  int numpts = data_size;

  spaces = new int[numpts];
  dist = new float[numpts];

/* find the distance between adjacent points, find maximum */
	ARR<float> dataA(dims);
	ARR<float> dataB(dims);
	ARR<float> maxmin(dims);

	for ( i = 0; i < dims; i++ )
	{
		maxmin[i] = (dim_max[i] - dim_min[i]) * (dim_max[i] - dim_min[i]);
	}

	for(i = 0;i < numpts-1;i++)   
	{
		GetData(dataA,i);
		GetData(dataB,i+1);
		dist[i] = comp_dist(dataA, dataB, maxmin, dims);
		if(dist[i] > maxdist)
			maxdist = dist[i];
	}

/* put spacing proportional to their distances, and calculate total space needs
*/
  total_space = 0;
  for(i = 0;i < numpts-1;i++)   {
    spaces[i] = 1 + MAX_SPACE * dist[i] / maxdist;
    total_space += spaces[i];
    }

  row_length = sqrt((double)total_space) + .5;  /* number of glyphs that will fit in
        a row, assuming a square canvas */

  blocksize = 1.0 / row_length;     /* max size of a glyph, assumes square canvas */
  halfblocksize = blocksize/2.0;            /* pos is where center of glyph is, so need offset */

/* comput the locations of the glyphs based on how much space is between them
*/
  index = 0;
  for(i = 0;i < numpts;i++)     {
    basex[i] = (index % row_length) * blocksize + halfblocksize;
    basey[i] = (index / row_length) * blocksize + halfblocksize;
    index += spaces[i];
    }
 /* now you can draw the glyphs at locations stored in pos */
}


//decide the base of every glyph according to their order in one dimension
void okcdata::Placement_Sort_Space_Glyph(int dimNumber, bool increase)
{
	SortStruct *sortArray = new SortStruct[data_size];
	
	int i, k;
	k = dimNumber;
	for ( i = 0; i < data_size; i++ )			
	{
		sortArray[i].number = i;
		sortArray[i].value = data_buf[k];
		k = k+all_dims;
	}

	if ( increase == true )
		qsort( sortArray, data_size, sizeof( SortStruct ), IncreaseSort );
	else 
		qsort( sortArray, data_size, sizeof( SortStruct ), DecreaseSort );

	int index;
	int size, halfsize;   /* size of a glyph, and half the size, to center
	position */
	float maxdist = 0.0;
	int row_length;       /* number of glyphs that will fit on a row */

	/* get memory for lists */
	int numpts = data_size;

	spaces = new int[numpts];
	dist = new float[numpts];

	/* find the distance between adjacent points, find maximum */
	ARR<float> dataA(dims);
	ARR<float> dataB(dims);
	ARR<float> maxmin(dims);

	for ( i = 0; i < dims; i++ )
	{
		maxmin[i] = (dim_max[i] - dim_min[i]) * (dim_max[i] - dim_min[i]);
	}

	for(i = 0;i < numpts-1;i++)   
	{
		GetData(dataA,sortArray[i].number);
		GetData(dataB,sortArray[i+1].number);
		dist[i] = comp_dist(dataA, dataB, maxmin, dims);
		if(dist[i] > maxdist)
			maxdist = dist[i];
	}

	/* put spacing proportional to their distances, and calculate total space needs 	*/
	total_space = 0;
	for(i = 0;i < numpts-1;i++)   
	{
		spaces[i] = 1 + MAX_SPACE * dist[i] / maxdist;
		total_space += spaces[i];
	}

	row_length = sqrt((double)total_space) + .5;  /* number of glyphs that will fit in
	a row, assuming a square canvas */

	blocksize = 1.0 / row_length;     /* max size of a glyph, assumes square canvas */
	halfblocksize = blocksize/2.0;            /* pos is where center of glyph is, so need offset */

	/* comput the locations of the glyphs based on how much space is between them	*/
	index = 0;
	for(i = 0;i < numpts;i++)     
	{
		basex[sortArray[i].number] = (index % row_length) * blocksize + halfblocksize;
		basey[sortArray[i].number] = (index / row_length) * blocksize + halfblocksize;
		index += spaces[i];
	}
 /* now you can draw the glyphs at locations stored in pos */

	delete sortArray;
	
}


/*************************************************************************************************/
//Principal component analysis -- derived

#define SIGN(a, b) ( (b) < 0 ? -fabs(a) : fabs(a) )

float **matrix(int n, int m)
{
	int i;
	float **mat;

	mat = new float *[n];
	
	for ( i = 0; i < n; i++ )
	{
		mat[i] = new float[m];
	}
	return mat;
}

void free_matrix(float** matrix, int n, int m)
{
	for ( int i = 0; i < n; i++ )
	{
		delete matrix[i];
	}
	delete matrix;
}

float *vector(int n)
{
	float* v;
	v = new float[n];
	return v;
}

void free_vector(float *v, int n)
{
	delete v;
}

//creat correlation matrix
void corcol(float **data, int n, int m, float **symmat)
{
	float eps = 0.005;
	float x, *mean, *stddev;
	int i, j, j1, j2;

	mean = vector(m);
	stddev = vector(m);

	// determine mean of column vectors of input data matrix
	for ( j = 0; j < m; j++ )
	{
		mean[j] = 0.0;
		for ( i = 0; i < n; i++ )
		{
			mean[j] += data[i][j];
		}
		mean[j] /= n;
	}

	//determine standard deviations of colunm vectors of data matrix
	for ( j = 0; j < m; j++ )
	{
		stddev[j] = 0.0;
		for ( i = 0; i < n; i++ )
		{
			stddev[j] += ( ( data[i][j] - mean[j] ) *
				           ( data[i][j] - mean[j] ) );
		}
		stddev[j] /= n;
		stddev[j] = sqrt(stddev[j]);
		/* the following is an inelegant but usual way to handle near-zero std. dev. values, which
		below would cause a zero-divide */
		if ( stddev[j] <= eps )
			stddev[j] = 1.0;
	}

	//center and reduce the column vectors
	for ( i = 0; i < n; i++ )
	{
		for ( j = 0; j < m; j++ )
		{
			data[i][j] -= mean[j];
			x = sqrt((float)n);
			x *= stddev[j];
			data[i][j] /= x;
		}
	}

	//calculate the m*m correlation matrix
	for ( j1 = 0; j1 < m-1; j1++ )
	{
		symmat[j1][j1] = 1.0;
		for ( j2 = j1+1; j2 < m; j2++ )
		{
			symmat[j1][j2] = 0.0;
			for ( i = 0; i < n; i++ )
			{
				symmat[j1][j2] += ( data[i][j1] * data[i][j2] );
			}
			symmat[j2][j1] = symmat[j1][j2];
		}
	}
	symmat[m-1][m-1] = 1.0;
	return;
}

//householder reduction of matrix a  to tridiagonal form
void tred2( float **a, int n, float *d, float *e)
{
	int  l, k, j, i;
	float scale, hh, h, g, f;

	for ( i = n-1; i >= 1; i-- )
	{
		l = i - 1; 
		h = scale = 0.0;
		if ( l > 0 )
		{
			for ( k = 0; k <= l; k++ )
				scale += fabs(a[i][k]);
			if ( fabs(scale) < 1e-8 )
				e[i] = a[i][l];
			else
			{
				for ( k = 0; k <= l; k++ )
				{
					a[i][k] /= scale;
					h += a[i][k] * a[i][k];
				}
				f = a[i][l];
				g = f>0 ? -sqrt(h) : sqrt(h);
				e[i] = scale * g;
				h -= f * g;
				a[i][l] = f - g;
				f = 0.0;
				for ( j = 0; j <= l; j++ )
				{
					a[j][i] = a[i][j] / h;
					g = 0.0;
					for ( k = 0; k <= j; k++ )
						g += a[j][k] * a[i][k];
					for ( k = j+1; k <= l; k++ )
						g += a[k][j] * a[i][k];
					e[j] = g / h;
					f += e[j] * a[i][j];
				}
				hh = f / ( h + h );
				for ( j = 0; j <= l; j++ )
				{
					f = a[i][j];
					e[j] = g = e[j] - hh * f;
					for ( k = 0; k <= j; k++ )
						a[j][k] -= ( f * e[k] + g * a[i][k] );
				}
			}
		}
		else 
			e[i] = a[i][l];
		d[i] = h;
	}

	d[0] = 0.0;
	e[0] = 0.0;
	for ( i = 0; i < n; i ++ )
	{
		l = i - 1; 
		if ( d[i] )
		{
			for ( j = 0; j <= l; j++ )
			{
				g = 0.0;
				for ( k = 0; k <= l; k++ )
					g += a[i][k]  * a[k][j];
				for ( k = 0; k <= l; k++ )
					a[k][j] -= g * a[k][i];
			}
		}
		d[i] = a[i][i];
		a[i][i] = 1.0;
		for ( j = 0; j <= l; j++ )
			a[j][i] = a[i][j] = 0.0;
	}
}

// tridiagonal QL algorithm -- implicit
bool tqli(float *d, float *e, int n, float **z)
{
	int m, l, iter, i, k;
	float s, r, p, g, f, dd, c, b;

	for ( i = 1; i < n; i++ )
		e[i-1] = e[i];
	e[n-1] = 0.0;
	for ( l = 0; l < n; l++ )
	{
		iter = 0; 
		do
		{
			for ( m = l; m < n-1; m++ )
			{
				dd = fabs(d[m]) + fabs(d[m+1]);
				if ( fabs(e[m]) < 1e-8 )
					break;
			}
			if (m != l )
			{
				if ( iter++ == 30 )
					return false;
				g = ( d[l+1] - d[l] ) / ( 2.0 * e[l] );
				r = sqrt( (g * g) + 1.0 );
				g = d[m] - d[l] + e[l] / ( g + SIGN(r, g) );
				s = c = 1.0;
				p = 0.0;
				for ( i = m-1; i >= l; i-- )
				{
					f = s * e[i];
					b = c * e[i];
					if ( fabs(f) >= fabs(g) )
					{
						c = g / f; 
						r = sqrt( ( c * c ) + 1.0 );
						e[i+1] = f * r;
						c *= ( s = 1.0/r );
					}
					else 
					{
						s = f / g;
						r = sqrt( (s * s) + 1.0 );
						e[i+1] = g * r;
						s *= ( c = 1.0/r );
					}
					g = d[i+1] - p;
					r = ( d[i] - g ) * s + 2.0 * c * b;
					p = s * r; 
					d[i+1] = g + p;
					g = c * r - b;
					for ( k = 0; k < n; k++ )
					{
						f = z[k][i+1];
						z[k][i+1] = s * z[k][i] + c * f;
						z[k][i] = c * z[k][i] - s * f;
					}
				}
				d[l] = d[l] - p;
				e[l] = g; 
				e[m] = 0.0;
			}
		}   while ( m != l );
	}
	return true;
}

void ReadDataFromBuffer( float **data, int n, int m )
{
	for ( int i = 0; i < n; i++ )
	{
	    ARR<float> datapoint(m);
        dataset->GetData(datapoint, i);
		for ( int j = 0; j < m; j++ )
		{
			data[i][j] = datapoint[j];
		}
	}
}

bool principal_Components_Analysis(int prin_num, float **prin, float *prin_max, float *prin_min)
{
	int n, m, i, j, k, k2;
	float **data, **symmat, **symmat2, *evals, *interm;
	
	n = dataset->data_size;
	m = dataset->dims;

	data = matrix(n, m);
	ReadDataFromBuffer(data, n, m);

	symmat = matrix(m, m); //symmat: correlation matrix
	corcol(data, n, m, symmat);	 //creat correlation matrix

/**** eigen-reduction ***********/

	evals = vector(m);   //evals: vector of eigenvalues
	interm = vector(m);   //interm: vector of intermediate
	symmat2 = matrix(m, m);  //duplication of correlation matrix
	for ( i = 0; i < m; i++ )
		for ( j = 0; j < m; j++ )
			symmat2[i][j] = symmat[i][j];

	tred2(symmat, m, evals, interm); //triangular decomposition
	
	if ( ! tqli(evals, interm, m, symmat) )
	{
		free_matrix(data, n, m); 
		free_matrix(symmat, m, m); 
		free_matrix(symmat2, m, m); 
		free_vector(evals, m); 
		free_vector(interm, m); 
		return false; //reduction of sym. trid. matrix
	}
/* evals now contains the eigenvalues, columns of symmat now contain the associated eigenvectors*/

	// Form projections of row-points on first three prin. components
	for ( i = 0; i < n; i++ )
	{
		for ( k = 0; k < prin_num; k++ )
		{
			prin[k][i] = 0.0;
			for ( k2 = 0; k2 < m; k2++ )
				prin[k][i] += data[i][k2] * symmat[k2][m-k-1];
		}
	}

	for ( i = 0; i < prin_num; i++ )
		prin_max[i] = prin_min[i] = prin[i][0];

	for ( i = 0; i < n; i++ )
	{
		for ( j = 0; j < prin_num; j++ )
		{
			if ( prin[j][i] > prin_max[j] )
				prin_max[j] = prin[j][i];
			if ( prin[j][i] < prin_min[j] )
				prin_min[j] = prin[j][i];
		}
	}

	free_matrix(data, n, m); 
	free_matrix(symmat, m, m); 
	free_matrix(symmat2, m, m); 
	free_vector(evals, m); 
	free_vector(interm, m); 
	return true;
}


void okcdata::Placement_Derived_Glyph(int XdimNumber, int YdimNumber)
{
	int prin_num;
	float **principal;
	float *principal_max;
	float *principal_min;
	char **principal_names;
	int i;

	if ( dataset->dims < 3 )
		prin_num = dims;
	else
		prin_num = 3;

	principal = new float *[prin_num];
	for ( i = 0; i < prin_num; i++ )
		principal[i] = new float[data_size];
	principal_max = new float[prin_num];
	principal_min = new float[prin_num];

	if ( !principal_Components_Analysis(prin_num, principal, principal_max, principal_min) )
	{
		for ( i = 0; i < prin_num; i++ )
			delete principal[i];
		delete principal;
		delete principal_max;
		delete principal_min;
		return;
	}

	int base;

	if ( XdimNumber < all_dims )
	{
		float xmin = all_dim_min[XdimNumber]; 
		float xpara = 1.0/(all_dim_max[XdimNumber] - xmin);
		
		base = 0;
		for ( i = 0; i < data_size; i++ )
		{
			basex[i] = (data_buf[base+XdimNumber] - xmin) * xpara;
			base += all_dims;
		}
	}
	else
	{
		XdimNumber -= all_dims;
		float xmin = principal_min[XdimNumber]; 
		float xpara = 1.0/(principal_max[XdimNumber] - xmin);
		
		for ( i = 0; i < data_size; i++ )
			basex[i] = (principal[XdimNumber][i] - xmin) * xpara;
	}
	
	if ( YdimNumber < all_dims )
	{
		float ymin = all_dim_min[YdimNumber]; 
		float ypara = 1.0/(all_dim_max[YdimNumber] - ymin);
		
		base = 0;
		for ( i = 0; i < data_size; i++ )
		{
			basey[i] = 1-(data_buf[base+YdimNumber] - ymin) * ypara;
			base += all_dims;
		}
	}
	else
	{
		YdimNumber -= all_dims;
		float ymin = principal_min[YdimNumber]; 
		float ypara = 1.0/(principal_max[YdimNumber] - ymin);
		
		for ( i = 0; i < data_size; i++ )
			basey[i] = 1-(principal[YdimNumber][i] - ymin) * ypara;
	}

	for ( i = 0; i < prin_num; i++ )
		delete principal[i];
	delete principal;
	delete principal_max;
	delete principal_min;
	
}
