////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty. 
//  In no event will the authors be held liable for any damages arising 
//  from the use of this software.
// 
//  Authors:
//      Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //
#include "glyphtool.H"
#include "XmdvTool.H"
#include "glyph.H"
#include "brush.H"
#include "ogl.H"

static int axis = -1;
static float mouse_angle;
static int resize_which = -1;

const float R = 0.25; // Center width

void RegisterGlyphToolCanvas(grCanvas *canvas)
{
    canvas->display_proc = glyphtoolRedrawCanvas;
    canvas->motion_proc = glyphtoolMotionNotify;
    canvas->buttonpress_proc = glyphtoolButtonPress;
    canvas->buttonrelease_proc = glyphtoolButtonRelease;
}

void glyphtoolRedrawCanvas(grCanvas *canvas)
{
    if (!dataset) return;
    
    // Setup window parameters.
    //
    int W, H;
    canvas->GetSize(&W,&H);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    const float BORDER = 0.25;
    const float L = 1 + R + BORDER;
    glViewport( 0, 0, W, H);
    glOrtho(-L,L,-L,L,-L,L);

    canvas->SetBackground(interface_color[BG]);
    canvas->Clear();

    // Precompute table of values
    //
    int N = dataset->dims;
    int i;
    float angle = 0;

    ARR<float> cosa(N), sina(N), lo(N), hi(N);
    
    for (i=0;i<N;i++) {
        cosa[i] = cos(angle);
        sina[i] = sin(angle);
        angle += 2.*M_PI*dataset->one_over_dims;
        lo[i] = (curr_brush->Min(i)-dataset->org_dim_min[i])/
                (dataset->org_dim_max[i] - dataset->org_dim_min[i]) + R;
        hi[i] = (curr_brush->Max(i)-dataset->org_dim_min[i])/
                (dataset->org_dim_max[i] - dataset->org_dim_min[i]) + R;
    }

    // Draw the brush
    //
    canvas->SetForeground(interface_color[BRUSH1+curr_brush->num]);

    ARR<Vec2> pts(4);
    pts[0].X = cosa[N-1] * hi[N-1];
    pts[1].X = cosa[N-1] * lo[N-1];

    pts[0].Y = sina[N-1] * hi[N-1];
    pts[1].Y = sina[N-1] * lo[N-1];

    for (i=0; i<N; i++ )
    {
        pts[3].X = cosa[i] * hi[i];
        pts[3].Y = sina[i] * hi[i];

        pts[2].X = cosa[i] * lo[i];
        pts[2].Y = sina[i] * lo[i];

        canvas->FillPolygon(pts);
        
        pts[0] = pts[3];
        pts[1] = pts[2];
    }

    // Draw the axes, labels
    //
    Vec2 size = canvas->GetDataFontSize("X");

    for (i=0; i<N; i++) 
    {
        canvas->SetForeground(interface_color[GRID1]);
        canvas->DrawLine(R*cosa[i],R*sina[i],(1+R)*cosa[i],(1+R)*sina[i]);

        float angle = i*2.*M_PI*dataset->one_over_dims;

        float dx;
        if (angle<M_PI) dx = -angle/M_PI;
        else dx = -(2*M_PI - angle)/M_PI;
        
        Vec2 pos;
        pos.X = (1.01+R)*cosa[i] + dx*BORDER*.8;
        pos.Y = (1.01+R)*sina[i] - (size.Y/2.) - (size.Y/2.) * sin(angle+M_PI);

        canvas->SetForeground(interface_color[TEXT]);
        canvas->DrawConfinedString(pos.X,pos.Y,dataset->names[i],BORDER);
    }

    // Draw the center loop
    //
    canvas->SetForeground(interface_color[GRID1]);
    for (i=1; i<N; i++) {
        canvas->DrawLine(R*cosa[i],R*sina[i],R*cosa[i-1],R*sina[i-1]);
    }
    canvas->DrawLine(R*cosa[N-1],R*sina[N-1],R*cosa[0],R*sina[0]);
}

static void glyphtoolResize(grCanvas *canvas, int x, int y)
{
    if (axis<0) return;

    // Map mouse location to data coordinates 
    Vec2 p = canvas->MapScreenToData(Vec2(x,y));

    // Find angle of axis 
    float angle = axis * (2*M_PI*dataset->one_over_dims);

    // Calculate distance from center 
    float dot = cos(angle)*p.X + sin(angle)*p.Y - R;
    if (dot>1) dot = 1;
    else if (dot<0) dot = 0;

    // Calculate new position and bound it 
    float new_pos = dataset->org_dim_min[axis]+dot*(dataset->org_dim_max[axis]-dataset->org_dim_min[axis]);
    new_pos = MIN(new_pos, dataset->org_dim_max[axis]);
    new_pos = MAX(new_pos, dataset->org_dim_min[axis]);

    float lo = curr_brush->Min(axis);
    float hi = curr_brush->Max(axis);

    if (resize_which<0) {
        if (fabs(new_pos-lo) < fabs(new_pos-hi)) {
            resize_which = 0;
        } else {
            resize_which = 1;
        }
    }

    if (resize_which==0) lo = MIN(new_pos,hi);
    else hi = MAX(new_pos, lo);

    curr_brush->SetMinMax(axis,lo,hi);
}

static void glyphtoolMove(grCanvas *canvas, int x, int y)
{
    if (axis<0) return;

    // Map mouse location to data coordinates 
    Vec2 p = canvas->MapScreenToData(Vec2(x,y));

    // Find angle of axis 
    float angle = axis * (2*M_PI*dataset->one_over_dims);

    // Calculate distance from center 
    float dot = cos(angle)*p.X + sin(angle)*p.Y - R;
    if (dot>1) dot = 1;
    else if (dot<0) dot = 0;

    float new_pos = dataset->org_dim_min[axis]+dot*(dataset->org_dim_max[axis]-dataset->org_dim_min[axis]); 
    new_pos = MIN(new_pos, dataset->org_dim_max[axis]-curr_brush->size[axis]/2.);
    new_pos = MAX(new_pos, dataset->org_dim_min[axis]+curr_brush->size[axis]/2.);

    curr_brush->pos[axis] = new_pos;
}

void glyphtoolMotionNotify(grCanvas *canvas, int x, int y)
{
    if (canvas->button[0]==true) {
        glyphtoolResize(canvas,x,y);
        canvas->Refresh();
    } else
    if (canvas->button[1]==true) {
        glyphtoolMove(canvas,x,y);
        canvas->Refresh();
    }
}

void glyphtoolButtonPress(grCanvas *canvas, int button, int x, int y)
{
    if (!dataset) return;
    
    Vec2 pick = canvas->MapScreenToData(Vec2(x,y));

    // Find the angle of the line from the center to the mouse 
    // position 
    mouse_angle = atan2(pick.Y, pick.X);

    // Add a half wedge to the angle 
    float angle = mouse_angle + ((2.*M_PI)/(float)dataset->dims)/2.;

    // Wrap angle around if it is negative 
    if (angle < 0.) angle += 2.*M_PI;

    // Map angle to an axis 
    axis = (int)(dataset->dims*(angle/(2.*M_PI)));

    // Sanity check 
    if (axis < 0 || axis >= dataset->dims) {
        axis = -1;
        return;
    }

    glyphtoolMotionNotify(canvas,x,y);
}

void glyphtoolButtonRelease(grCanvas *canvas, int button, int x, int y)
{
    if (!dataset || axis<0) return;

    appCanvasBrushHasChanged(Togl_Interp(canvas->GetWidget()));
    appRefreshAllCanvas(Togl_Interp(canvas->GetWidget()));
    axis = -1;
    resize_which = -1;
}
