////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//      Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //
//
// This file contains all routines associated with
// the multiple selection of points.
//

#include "multi_select.H"
#include "XmdvTool.H"
#include "brush.H"
#include "brush_oper.H"
#include "ogl.H"
#include "ARR.H"

const int MAX_MULTI_SEL = 2048;
bool multi_select = false;
static ARR<int> multi_sel_pts(MAX_MULTI_SEL);
static int num_multi_sel = 0;

//
// Define a new brush from the multiple
// selected points
///
static void multiSelectDefineBrush()
{
    if (!dataset || num_multi_sel<=0) return;

    int i,j;
    ARR<float> brush_mins(dataset->dims), brush_maxes(dataset->dims);
    ARR<float> data(dataset->dims);

    // Set starting value for mins/maxes
    for (i=0; i<dataset->dims; i++)
    {
        brush_mins[i] = HUGE_VAL;
        brush_maxes[i] = -HUGE_VAL;
    }

    // Define the brush min/max values
    for (i=0; i<num_multi_sel; i++)
    {
        dataset->GetData(data, multi_sel_pts[i]);

        for (j=0; j<dataset->dims; j++)
        {
            brush_mins[j] = MIN(brush_mins[j], data[j]);
            brush_maxes[j] = MAX(brush_maxes[j], data[j]);
        }
    }

    // Create the new brush from min/max values and redraw
    for (i=0; i<dataset->dims; i++)
    {
        curr_brush->pos[i] = (brush_mins[i] + brush_maxes[i]) / 2.;
        curr_brush->size[i] = brush_maxes[i] - brush_mins[i];

        // add on 1% of the min/max range for roundoff error
        curr_brush->size[i] += .01*(dataset->dim_max[i] - dataset->dim_min[i]);
    }

    BrushOperMarkDataset();
}

//
// Action that toggles the multi_select flag
// when a shift key is pressed or released.
//
void multiSelectSetState(grCanvas *canvas, bool state)
{
    multi_select = state;

    if (state) {
        num_multi_sel = 0;
        canvas->SetCursor("hand2");
    } else {
        // Create the brush from the points
        //
        if (num_multi_sel != 0) {
            multiSelectDefineBrush();
            appRefreshAllCanvas(Togl_Interp(canvas->GetWidget()));
        }
        canvas->SetDefaultCursor();
    }
}

//
// Add a point to the list of points in the
// multi-select list.  If point already exists,
// will not add it.  Returns 0 on success and 1
// on error.
//
bool multiSelectAddPoint(int point)
{
    // check if full
    if (num_multi_sel == MAX_MULTI_SEL) return false;

    // check if point is already there
    for (int i=0; i<num_multi_sel; i++)
        if (multi_sel_pts[i] == point)
            return false;

    // add the point
    multi_sel_pts[num_multi_sel++] = point;

    return true;
}

//
// Delete a point from the multi-select list.
// No operation performed if point is not in
// the list.  Returns 0 on success or 1 on error.
//
bool multiSelectDeletePoint(int point)
{
    // check if point is already there
    int i;
    for (i=0; i<num_multi_sel; i++)
        if (multi_sel_pts[i] == point)
            break;

    if (i == num_multi_sel) // if it wasn't there
        return false;

    // delete the point
    for (int j=i+1; j<num_multi_sel; j++)
        multi_sel_pts[j-1] = multi_sel_pts[j];

    num_multi_sel--;
    return true;
}

//
// Given a point, return TRUE if the point is
// in the multi-select list.
//
bool multiSelectIsCovered(int point)
{
   for (int i=0; i<num_multi_sel; i++)
        if (multi_sel_pts[i] == point)
            return true;

    return false;
}
