////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//      Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //
//
// parcoord.c - function to plotting data using parallel coordinates for
// XmdvTool
//
// Display data using parallel coordinates.  Each dimensional axis is layed
// out in a vertical line, evenly spaced between dimensions.  The top and
// bottom of each line coincide with the maximum and minimum for that
// dimension.  Each data point creates a polyline across these axes.  Labels
// are printed along the top of each dimension, and brush coverage display
// consists of a single shaded region across the drawing area.
//

#include "parcoord.H"
#include "XmdvTool.H"
#include "ogl.H"
#include "brush.H"
#include "brush_oper.H"
#include "magnify.H"
#include "multi_select.H"


//
// Coordinate mappings in parallel coordinates from data space
// to normalized screen space
//
#define XMAP(I) (((I)+.5) * dataset->one_over_dims)
#define YMAP(X,Y) (1.0-((Y) - dataset->dim_min[X]) / (dataset->dim_max[X] - dataset->dim_min[X]))
#define YINVMAP(X,Y) (((1.0-(Y)) * (dataset->dim_max[X] - dataset->dim_min[X])) + dataset->dim_min[X])

int maskBrushedData = 0;
int maskUnbrushedData = 0;
int maskBrush = 0;
int brushPaint = 0;
//jing this func Rewrite in hier
void RegisterParcoordCanvas(grCanvas *canvas)
{
    canvas->display_proc = do_par;
    canvas->reshape_proc = grCanvas::grResizeMainCanvas;
    canvas->motion_proc = InputParcoordMouseMotion;
    canvas->buttonpress_proc = InputParcoordButtonPress;
    canvas->buttonrelease_proc = InputParcoordButtonRelease;
    canvas->keypress_proc = appCanvasInputKeyPress;
    canvas->keyrelease_proc = appCanvasInputKeyRelease;
}

//
// Draw the brush boundary in the parallel coordinate view.
//
//jing this func is called by a func used by hier
static void DrawParcoordBrushBoundRamp(grCanvas *canvas, Brush *brush, unsigned long color1, unsigned long color2)
{
    // set the appropriate color
    //
    ARR<Vec2> poly(dataset->dims);

    float factor = 0.1;
    RGBt c1(color1), c2(color2);
    c2.R = factor*c1.R + (1-factor)*c2.R;
    c2.G = factor*c1.G + (1-factor)*c2.G;
    c2.B = factor*c1.B + (1-factor)*c2.B;

    canvas->BeginDraw();

    glBegin(GL_QUAD_STRIP);
    int i;
    for (i=0; i<dataset->dims; i++)
    {
        float x = XMAP(i);
        float y1 = YMAP(i, brush->RampMax(i) );
        float y2 = YMAP(i, brush->Max(i) );

        glColor3f(c2.R,c2.G,c2.B);
        glVertex2f(x, y1);
        glColor3f(c1.R,c1.G,c1.B);
        glVertex2f(x, y2);
    }
    glEnd();

    glBegin(GL_QUAD_STRIP);
    for (i=0; i<dataset->dims; i++)
    {
        float x = XMAP(i);
        float y1 = YMAP(i, brush->RampMin(i) );
        float y2 = YMAP(i, brush->Min(i) );

        glColor3f(c2.R,c2.G,c2.B);
        glVertex2f(x, y1);
        glColor3f(c1.R,c1.G,c1.B);
        glVertex2f(x, y2);
    }
    glEnd();

    canvas->EndDraw();
}

//
// Draw the brush boundary in the parallel coordinate view.
//
//jing this func is used by a func used by canvas->.. _proc
static void DrawParcoordBrushBoundRampOutline(grCanvas *canvas, Brush *brush)
{
    // set the appropriate color
    //
    ARR<Vec2> poly(dataset->dims);

    int i;
    for (i=0; i<dataset->dims; i++)
    {
        poly[i].X = XMAP(i);
        poly[i].Y = YMAP(i, brush->RampMax(i) );
    }

    canvas->SetXorMode(True);
    canvas->DrawLines(poly);

    for (i=0; i<dataset->dims; i++)
    {
        poly[i].X = XMAP(i);
        poly[i].Y = YMAP(i, brush->RampMin(i) );
    }

    canvas->DrawLines(poly);
    canvas->SetXorMode(False);
}

//
// Draw an average value on the parallel
// coordinate display.
//
//jing I will not use this func temporarily
void DrawParcoordAverage(grCanvas *canvas, const ARR<float> &data)
{
    // draw polyline
    ARR<Vec2> poly(dataset->dims);

    for (int j = 0;j < dataset->dims;j++) {
        poly[j].X = XMAP(j);
        poly[j].Y = YMAP(j,data[j]);
    }

    canvas->DrawLines(poly);
}

//
// Draw standard vertical uniformly spaced parallel
// coordinate axes.
//
//jing this func is used by hier
void DrawParcoordAxes(grCanvas *canvas)
{
    float dimwidth = 1.0/dataset->dims;

    // Draw the vertical axes
    //
    canvas->SetForeground(interface_color[GRID1]);

    int i;
    for (i=0; i<dataset->dims; i++) {
        float x = XMAP(i);
        canvas->DrawLine(x, 0, x, 1.0);
    }

    canvas->SetForeground(interface_color[TEXT]);

    Vec2 font_size = canvas->GetDataFontSize("X");
    float max_str_size = dimwidth - font_size.X;

    for (i=0; i<dataset->dims; i++)
    {
        float x = XMAP(i) + font_size.X;
        canvas->DrawConfinedString(x,font_size.Y,dataset->names[i],max_str_size);

        // Draw the MAX values of the dimension
        char buf[50];
        sprintf(buf, "%.02f", dataset->dim_max[i]);
        canvas->DrawConfinedString(x,-.5*font_size.Y,buf,max_str_size);

        // Draw the MIN values of the dimension
        sprintf(buf, "%.02f", dataset->dim_min[i]);
        canvas->DrawConfinedString(x,1.0-font_size.Y,buf,max_str_size);
    }
}

//
// Draw brush set in parallel
// coordinate space.
//
//jing this func is used by hier
void DrawParcoordBrush(grCanvas *canvas, Brush *brush)
{
    canvas->SetForeground( interface_color[BRUSH1+brush->num]);
    canvas->BeginDraw();
    glBegin(GL_QUAD_STRIP);

    for (int j=0; j<dataset->dims; j++)
    {
        Vec2 lo, hi;

        // x positions are fixed y positions normalized over axis height

        // top points
        hi.X = XMAP(j);
        hi.Y = YMAP(j, brush->Max(j) );

        // bottom points
        lo.X = XMAP(j);
        lo.Y = YMAP(j, brush->Min(j) );

        glVertex2f(lo.X,lo.Y);
        glVertex2f(hi.X,hi.Y);
    }

    glEnd();
    canvas->EndDraw();

    // draw boundary coverage if necessary
    //
    if (brush->bound_type != Brush::STEP)
        DrawParcoordBrushBoundRamp(canvas,brush,interface_color[BRUSH1+brush->num],interface_color[BG]);
}

//jing this func is used by hier
void DrawParcoordBrushes(grCanvas *canvas)
{
    // if brush coverage is to be displayed, compute polyline by going first
    // through the top edges of the brush and then through the bottom edges
    //
    for (int brnum=0; brnum<MAXBRUSH; brnum++) {
        Brush *brush = &brushes[brnum];

        // if this brush is not being displayed, skip it
        if (brush->display == false) continue;

        DrawParcoordBrush(canvas,brush);
    }
}

//jing this func is Rewrited in hier
void do_par(grCanvas *canvas)
{
    if (!dataset) return;

    canvas->SetTopLeftOrigin();
    grCanvas::grResizeMainCanvas(canvas);

    canvas->SetBackground(interface_color[BG]);
    canvas->Clear();

	if ( maskBrush == 0 )
	    DrawParcoordBrushes(canvas);

    ARR<Vec2> poly(dataset->dims);
    ARR<float> data(dataset->dims);

    // first draw all points in the data color
    //
    RGBt c(interface_color[DATA]);
    glColor3f(c.R, c.G, c.B );

    glLineWidth(DSIZE);
    glShadeModel(GL_FLAT);  // Because each line is the same color.

    int i;
	if ( maskUnbrushedData == 0 )
	{
		for (i=0; i<dataset->data_size; i++)
		{
			// First time through only paint regular data
			//
			if (dataset->mark[i] == okcdata::DISABLED) continue;
			if (dataset->mark[i] != okcdata::REGULAR) continue;

			dataset->GetData(data, i);
			glBegin(GL_LINE_STRIP);

			for (int j=0; j<dataset->dims; j++) {
				glVertex2f( XMAP(j), YMAP(j,data[j]) );
			}
			glEnd();
		}
	}
    // now paint all points covered by an operation
    //
	if ( maskBrushedData == 0 )
	{
		for (i=0; i<dataset->data_size; i++)
		{
			if (dataset->mark[i] == okcdata::DISABLED) continue;
			if (dataset->mark[i] == okcdata::REGULAR) continue;

			dataset->GetData(data, i);

			glColor3f(dataset->color[i].R, dataset->color[i].G, dataset->color[i].B );
			glBegin(GL_LINE_STRIP);

			double temp = dataset->dim_max[3];
			for (int j=0; j<dataset->dims; j++) {
				glVertex2f( XMAP(j), YMAP(j,data[j]) );
			}
			glEnd();
		}
	}
    // Draw average operations
    for (int op=0; op<num_oper; op++)
    {
        // Check if this operation has OPER_AVERAGE set */
        if (!(brush_ops[op].oper & AVERAGE)) continue;

        ARR<float> average(dataset->dims);
        if (BrushOperCompAverage(op, average)==0) continue;

        glLineWidth(3.0);
        canvas->SetForeground(interface_color[TEXT]);
        DrawParcoordAverage(canvas,average);
    }

    glLineWidth(1.);
    glShadeModel(GL_SMOOTH);
    DrawParcoordAxes(canvas);
}

//
// Redraw a single polyline in the the parallel
// coordinate display.
//
static void RedrawSingleParCoord(grCanvas *canvas, int point)
{
    ARR<float> data(dataset->dims);
    ARR<Vec2> poly(dataset->dims);

    dataset->GetData(data, point);

    for (int j = 0;j < dataset->dims;j++) {
        poly[j].X = XMAP(j);
        poly[j].Y = YMAP(j,data[j]);
    }

    canvas->DrawLines(poly);
}

//
// Set of local variables to maintain
// the current state of interaction
//
static int axis = -1;
static bool xor_needs_undo = false;

enum INPUT_MODE { RESIZE_BRUSH_TOP, RESIZE_BRUSH_BOTTOM, TRANSLATE_BRUSH, RESIZE_RAMP, MULTI_SELECT, CANVAS_PAN, NONE };

static INPUT_MODE input_mode = NONE;

//jing this func is used by canvas->.._proc
static void InputParcoordResizeBrush(grCanvas *canvas, int sx, int sy)
{
    if (axis < 0 || axis >= dataset->dims) return;

    Vec2 p = canvas->MapScreenToData( Vec2(sx,sy) );

    // Clamp y coordinate to within dimension limits
    float y = YINVMAP(axis,p.Y);
    y = MAX(y,dataset->dim_min[axis]);
    y = MIN(y,dataset->dim_max[axis]);

    canvas->SetXorMode(true);

    float mx = XMAP(axis);

    switch (input_mode) {
    case RESIZE_BRUSH_TOP: {

        float ly = curr_brush->Min(axis);
        float hy = MAX(y,ly);
        p.Y = YMAP(axis,hy);

        curr_brush->SetMinMax(axis,ly,hy);

        // Resize top
        //
        if (axis>0) {
            float lx = XMAP(axis-1);
            float lhy = YMAP(axis-1,curr_brush->Max(axis-1));
            canvas->DrawLine( lx, lhy, mx, p.Y);
        }

        if (axis<dataset->dims-1) {
            float rx = XMAP(axis+1);
            float rhy = YMAP(axis+1,curr_brush->Max(axis+1));
            canvas->DrawLine( rx, rhy, mx, p.Y);
        }

        } break;

    case RESIZE_BRUSH_BOTTOM: {

        float hy = curr_brush->Max(axis);
        float ly = MIN(y,hy);
        p.Y = YMAP(axis,ly);

        curr_brush->SetMinMax(axis,ly,hy);

        // Resize bottom
        //
        if (axis>0) {
            float lx = XMAP(axis-1);
            float lly = YMAP(axis-1,curr_brush->Min(axis-1) );
            canvas->DrawLine( lx, lly, mx, p.Y);
        }

        if (axis<dataset->dims-1) {
            float rx = XMAP(axis+1);
            float rly = YMAP(axis+1,curr_brush->Min(axis+1) );
            canvas->DrawLine( rx, rly, mx, p.Y);
        }
        } break;

    default:
        break;
    }

    canvas->SetXorMode(false);
    xor_needs_undo = true;
}

//jing this func is used by canvas->.. _proc
static void InputParcoordMoveBrush(grCanvas *canvas, int sx, int sy)
{
    if (axis < 0 || axis >= dataset->dims) return;

    Vec2 p = canvas->MapScreenToData( Vec2(sx,sy) );

    canvas->SetXorMode(true);

    float mx = XMAP(axis);
    float y = YINVMAP(axis,p.Y);
    y = MIN(y,dataset->dim_max[axis]-curr_brush->size[axis]*.5);
    y = MAX(y,dataset->dim_min[axis]+curr_brush->size[axis]*.5);

    float mly = YMAP(axis,y-curr_brush->size[axis]*.5);
    float mhy = YMAP(axis,y+curr_brush->size[axis]*.5);

    if (axis>0) {
        float lx = XMAP(axis-1);
        float lly = YMAP(axis-1,curr_brush->Min(axis-1) );
        float lhy = YMAP(axis-1,curr_brush->Max(axis-1) );
        canvas->DrawLine( lx, lly, mx, mly);
        canvas->DrawLine( lx, lhy, mx, mhy);
    }

    if (axis<dataset->dims-1) {
        float rx = XMAP(axis+1);
        float rly = YMAP(axis+1,curr_brush->Min(axis+1) );
        float rhy = YMAP(axis+1,curr_brush->Max(axis+1) );
        canvas->DrawLine( rx, rly, mx, mly);
        canvas->DrawLine( rx, rhy, mx, mhy);
    }

    canvas->SetXorMode(false);

    curr_brush->pos[axis] = y;
    xor_needs_undo = true;
}

//jing this func is used by canvas->.. _proc
static void InputParcoordMultiSelect(grCanvas *canvas, int sx, int sy)
{
    if (axis < 0 || axis >= dataset->dims) return;
    if (!multi_select) return;

    ARR<float> data(dataset->dims);

    // Find all points within circle distance of mouse
    //
    for (int i=0; i<dataset->data_size; i++)
    {
        dataset->GetData(data, i);

        float y = YMAP(axis,data[axis]);

        Vec2 p = canvas->MapDataToScreen(Vec2(XMAP(axis),y));

        if (fabs(p.Y - sy) < multisel_radius)
        {
            multiSelectAddPoint(i);

            glDrawBuffer(GL_FRONT);
            canvas->SetForeground(interface_color[PAINT]);
            RedrawSingleParCoord(canvas, i);
            glDrawBuffer(GL_BACK);
        }
    }
}

//jing this func is used by canvas->.._proc
static void InputParcoordBoundRampResize(grCanvas *canvas, int sx, int sy)
{
    if (axis < 0 || axis >= dataset->dims) return;

    Vec2 p = canvas->MapScreenToData( Vec2(sx,sy) );

    float y = YINVMAP(axis,p.Y);

    if (y > curr_brush->pos[axis])
    {
        curr_brush->bound_offset[axis] = y - curr_brush->Max(axis);
        curr_brush->bound_offset[axis] = MAX(curr_brush->bound_offset[axis], 0.0);
    }
    else
    {
        curr_brush->bound_offset[axis] = curr_brush->Min(axis) - y;
        curr_brush->bound_offset[axis] = MAX(curr_brush->bound_offset[axis], 0.0);
    }

    DrawParcoordBrushBoundRampOutline(canvas,curr_brush);

    xor_needs_undo = true;
}

//jing this func is used by canvas->.. _proc
static void ParcoordCaption(grCanvas *canvas, int x, int y)
{
    Vec2 p = canvas->MapScreenToData( Vec2(x,y) );
    int datax = int(p.X * dataset->dims);
    datax = CLAMP(datax,0,dataset->dims-1);
    float datay = YINVMAP(datax,p.Y);

    char caption[100];

    if (dataset->nominal_fields[datax] != 0) {
//        int value = (int) rint(datay);
//jing
		int value = (int)(datay+0.5);
//jing
        const char *string = dataset->nominal_fields[datax]->MapValueToString(value);
        sprintf(caption,"%s=%s",dataset->names[datax],string);
    } else {
        sprintf(caption,"%s=%.2f",dataset->names[datax],datay);
    }

    appShowCaption(Togl_Interp(canvas->GetWidget()),caption);
}

//jing canvas->motion_proc, Rewrite in hier
void InputParcoordMouseMotion(grCanvas *canvas, int x, int y)
{
    if (!dataset) return;

    ParcoordCaption(canvas, x, y);
    static int prev_sx, prev_sy;

    switch (input_mode) {

        case CANVAS_PAN:
            appCanvasMousePan(canvas);
            return;

        case RESIZE_RAMP:
            if (xor_needs_undo) InputParcoordBoundRampResize(canvas,prev_sx,prev_sy);
            InputParcoordBoundRampResize(canvas,x,y);
            break;

        case RESIZE_BRUSH_TOP:
        case RESIZE_BRUSH_BOTTOM:
            if (xor_needs_undo) InputParcoordResizeBrush(canvas,prev_sx,prev_sy);
            InputParcoordResizeBrush(canvas,x,y);
            break;

        case TRANSLATE_BRUSH:
            if (xor_needs_undo) InputParcoordMoveBrush(canvas,prev_sx,prev_sy);
            InputParcoordMoveBrush(canvas,x,y);
            break;

        case MULTI_SELECT:
            InputParcoordMultiSelect(canvas,x,y);
            return;

        default: break;
    }

    prev_sx = x;
    prev_sy = y;
}

//jing canvas->buttonpress_proc, Rewrite in hier
void InputParcoordButtonPress(grCanvas *canvas, int button, int x, int y)
{
    if (!dataset) return;

    input_mode = NONE;

    if (canvas->button[2] == true) {
        input_mode = CANVAS_PAN;
        canvas->SetHandCursor();
        return;
    }

    Vec2 p = canvas->MapScreenToData( Vec2(x,y) );
    axis = int(p.X * dataset->dims);

    if (axis < 0 || axis >= dataset->dims) {
        axis = -1;
        return;
    }

    if (canvas->shiftkey && button == 0) {
        input_mode = MULTI_SELECT;
        InputParcoordMultiSelect(canvas,x,y);
        return;
    } else
    if (canvas->controlkey && curr_brush->bound_type == Brush::RAMP && button == 0) {
        input_mode = RESIZE_RAMP;
        InputParcoordBoundRampResize(canvas,x,y);
        return;
    }

    assert(xor_needs_undo==false);

    if (canvas->button[0] == true) {

        float lo = YMAP(axis,curr_brush->Min(axis));
        float hi = YMAP(axis,curr_brush->Max(axis));
        if ( (hi-p.Y) > (p.Y - lo) ) {
            input_mode = RESIZE_BRUSH_TOP;
        } else {
            input_mode = RESIZE_BRUSH_BOTTOM;
        }
    } else
    if (canvas->button[1] == true) {
        input_mode = TRANSLATE_BRUSH;
    }

    InputParcoordMouseMotion(canvas,x,y);
}

//jing canvas->buttonrelease_proc, Rewrite in hier
void InputParcoordButtonRelease(grCanvas *canvas, int button, int x, int y)
{
    if (!dataset) return;

    switch (input_mode) {

        case CANVAS_PAN:
            canvas->SetDefaultCursor();
            break;

        case MULTI_SELECT:
            break;

        case RESIZE_RAMP:
        case RESIZE_BRUSH_TOP:
        case RESIZE_BRUSH_BOTTOM:
        case TRANSLATE_BRUSH:
            BrushOperMarkDataset();
            canvas->SetDefaultCursor();
            appRefreshAllCanvas(Togl_Interp(canvas->GetWidget()));
            break;

        default: break;
    }

    input_mode = NONE;
    xor_needs_undo = false;
}
