////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//      Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //
//
// scatterplot.c - function to plot data using scatterplots for XmdvTool
//
// Generate scatterplots by dividing drawing area into N by N grid and draw
// each data sample N * N times.  Draw grids, vertical and horizontal labels,
// and grids.  Redraw bits under brush if state set to ON
//

#include "scatter.H"
#include "parcoord.H"
#include "ogl.H"
#include "XmdvTool.H"
#include "brush.H"
#include "brush_oper.H"
#include "multi_select.H"
#include <string.h>

//
// Various coordinate mappings between screen space
// and normalized data space
//
#define XMAP(DIM,VAL) (DIM*dataset->one_over_dims+(((VAL) - dataset->dim_min[DIM]) * dataset->one_over_dims / (dataset->dim_max[DIM] - dataset->dim_min[DIM])))
#define YMAP(DIM,VAL) (DIM*dataset->one_over_dims+((dataset->dim_max[DIM] - (VAL)) * dataset->one_over_dims / (dataset->dim_max[DIM] - dataset->dim_min[DIM])))

#define XINVMAP(DIM,VAL) (((VAL)-(DIM)*dataset->one_over_dims) * (dataset->dim_max[DIM] - dataset->dim_min[DIM]) * dataset->dims + dataset->dim_min[DIM])
#define YINVMAP(DIM,VAL) (dataset->dim_max[DIM] - ((VAL)-(DIM)*dataset->one_over_dims) * (dataset->dim_max[DIM] - dataset->dim_min[DIM]) * dataset->dims)

void RegisterScatterCanvas(grCanvas *canvas)
{
    canvas->display_proc = do_scatter;
    canvas->reshape_proc = grCanvas::grResizeMainCanvas;
    canvas->motion_proc = InputScatterMouseMotion;
    canvas->buttonpress_proc = InputScatterButtonPress;
    canvas->buttonrelease_proc = InputScatterButtonRelease;
    canvas->keypress_proc = appCanvasInputKeyPress;
    canvas->keyrelease_proc = appCanvasInputKeyRelease;
}

//
// Redraw a single point on the scatterplot
// display.  Useful for multi-sel.
//
//jing this func is used by canvas->.. _proc
static void RedrawSingleScatter(grCanvas *canvas, int point)
{
//jing get data
    ARR<float> data(dataset->dims);
    dataset->GetData(data, point);

//jing set point size
    glPointSize((DSIZE+1)*canvas->zoom.X);

//jing draw a point in scatterplot matrix ( in dim*dim plots )
    canvas->BeginDraw();
    glBegin(GL_POINTS);

    for (int j = 0;j < dataset->dims;j++)
        for (int k = 0;k < dataset->dims;k++)
            glVertex2f( XMAP(j,data[j]), YMAP(k,data[k]) );

    glEnd();
//jing reset point size
    glPointSize(1.0);
}

//
// Draw an average value on the scatterplot
// display.
//
//jing I will not use this func temporarily
static void DrawScatterAverage(grCanvas *canvas, const ARR<float> &data)
{
    glPointSize(3*(DSIZE+1)*canvas->zoom.X);
    canvas->BeginDraw();
    glBegin(GL_POINTS);
    for (int j=0; j<dataset->dims; j++)
        for (int k=0; k<dataset->dims; k++)
            glVertex2f( XMAP(j,data[j]), YMAP(k,data[k]) );
    glEnd();
//jing why not in the last func? 
    canvas->EndDraw();
    glPointSize(1.0);
}


//
// Draw the brush boundary in the scatterplot
// view.
//
//jing this func is used by a func used by canvas->.. _proc
static void DrawScatterBrushBoundRampOutline(grCanvas *canvas, Brush *brush)
{
    canvas->SetXorMode(true);

    for (int j=0; j<dataset->dims; j++) {
        for (int k=0; k<dataset->dims; k++)
        {
            Vec2 p, q;

            p.X = brush->RampMin(j);
            p.Y = brush->RampMax(k);

            p.X = CLAMP(p.X, dataset->dim_min[j], dataset->dim_max[j]);
            p.Y = CLAMP(p.Y, dataset->dim_min[k], dataset->dim_max[k]);

            q.X = brush->RampMax(j);
            q.Y = brush->RampMin(k);

            q.X = CLAMP(q.X, dataset->dim_min[j], dataset->dim_max[j]);
            q.Y = CLAMP(q.Y, dataset->dim_min[k], dataset->dim_max[k]);

            p.X = XMAP(j, p.X);
            p.Y = YMAP(k, p.Y);
            q.X = XMAP(j, q.X);
            q.Y = YMAP(k, q.Y);

            canvas->DrawRectangle(p.X, p.Y, q.X-p.X, q.Y-p.Y);
        }
    }
    canvas->SetXorMode(false);
}	

//jing this func is called by a func used by hier
//jing draw the ramp around the brush
void DrawScatterBrushBoundRamp(grCanvas *canvas, Brush *brush, unsigned long color1, unsigned long color2)
{
    const float factor = 0.1;
    RGBt c1(color1), c2(color2);
    c2.R = factor*c1.R + (1-factor)*c2.R;
    c2.G = factor*c1.G + (1-factor)*c2.G;
    c2.B = factor*c1.B + (1-factor)*c2.B;

    canvas->BeginDraw();

    for (int j=0; j<dataset->dims; j++) {
        for (int k=0; k<dataset->dims; k++)
        {
            Vec2 p0,p1,q0,q1;

            p0.X = brush->Min(j);
            p0.Y = brush->Max(k);
            p1.X = p0.X - brush->bound_offset[j];
            p1.Y = p0.Y + brush->bound_offset[k];

            p1.X = CLAMP(p1.X, dataset->dim_min[j], dataset->dim_max[j]);
            p1.Y = CLAMP(p1.Y, dataset->dim_min[k], dataset->dim_max[k]);

            q0.X = brush->Max(j);
            q0.Y = brush->Min(k);
            q1.X = q0.X + brush->bound_offset[j];
            q1.Y = q0.Y - brush->bound_offset[k];

            q1.X = CLAMP(q1.X, dataset->dim_min[j], dataset->dim_max[j]);
            q1.Y = CLAMP(q1.Y, dataset->dim_min[k], dataset->dim_max[k]);

            p0.X = XMAP(j, p0.X);
            p0.Y = YMAP(k, p0.Y);
            p1.X = XMAP(j, p1.X);
            p1.Y = YMAP(k, p1.Y);
            q0.X = XMAP(j, q0.X);
            q0.Y = YMAP(k, q0.Y);
            q1.X = XMAP(j, q1.X);
            q1.Y = YMAP(k, q1.Y);

            glBegin(GL_QUAD_STRIP);

            glColor3f(c1.R,c1.G,c1.B);
            glVertex2f(p0.X,p0.Y);
            glColor3f(c2.R,c2.G,c2.B);
            glVertex2f(p1.X,p1.Y);
            glColor3f(c1.R,c1.G,c1.B);
            glVertex2f(p0.X,q0.Y);
            glColor3f(c2.R,c2.G,c2.B);
            glVertex2f(p1.X,q1.Y);
            glColor3f(c1.R,c1.G,c1.B);
            glVertex2f(q0.X,q0.Y);
            glColor3f(c2.R,c2.G,c2.B);
            glVertex2f(q1.X,q1.Y);
            glColor3f(c1.R,c1.G,c1.B);
            glVertex2f(q0.X,p0.Y);
            glColor3f(c2.R,c2.G,c2.B);
            glVertex2f(q1.X,p1.Y);
            glColor3f(c1.R,c1.G,c1.B);
            glVertex2f(p0.X,p0.Y);
            glColor3f(c2.R,c2.G,c2.B);
            glVertex2f(p1.X,p1.Y);

            glEnd();
        }
    }
    canvas->EndDraw();
}

//
// Given x and y dimensions, coordinates of the
// brush center, and brush size in each dimesnion
// draw the corresponding brush outline in xor mode
// on the scatterplot display.
//
//jing this func is used by funcs used by canvas->.. _proc
static void DrawScatterBrushBox(grCanvas *canvas, int dimx, int dimy, float data_x, float data_y,
             float size_x, float size_y)
{
    Vec2 p, q;

    // lower left
    p.X = XMAP(dimx, data_x - size_x/2);
    p.Y = YMAP(dimy, data_y - size_y/2);

    // upper right
    q.X = XMAP(dimx, data_x + size_x/2);
    q.Y = YMAP(dimy, data_y + size_y/2);

    canvas->SetXorMode(true);
    canvas->DrawRectangle(p.X,p.Y,q.X-p.X,q.Y-p.Y);
    canvas->SetXorMode(false);
}

//jing this func is Rewrited in hier
void do_scatter(grCanvas *canvas)
{
//jing common sentences
    if (!dataset) return;

    canvas->SetTopLeftOrigin();
    grCanvas::grResizeMainCanvas(canvas);

    canvas->SetBackground(interface_color[BG]);
    canvas->Clear();

    canvas->SetForeground(interface_color[TEXT]);
//jing draw brushes
    // Default box size just divides up drawing area
    // if brush coverage is on, draw regions covered by brush in outline color
    //
    if ( maskBrush == 0 )
	{
		for (int br=0; br<MAXBRUSH; br++)
		{
			Brush *brush = &brushes[br];

			// if this brush is not being displayed, skip it
			if (brush->display == false) continue;

			// set the appropriate brush color
			canvas->SetForeground(interface_color[BRUSH1+br]);
			canvas->SetXorMode(false);

			for (int j=0; j<dataset->dims; j++) {
				for(int k=0; k<dataset->dims; k++)
				{
					Vec2 p, q;

					p.X = brush->Min(j);
					p.Y = brush->Max(k);

					p.X = CLAMP(p.X, dataset->dim_min[j], dataset->dim_max[j]);
					p.Y = CLAMP(p.Y, dataset->dim_min[k], dataset->dim_max[k]);

					q.X = brush->Max(j);
					q.Y = brush->Min(k);

					q.X = CLAMP(q.X, dataset->dim_min[j], dataset->dim_max[j]);
					q.Y = CLAMP(q.Y, dataset->dim_min[k], dataset->dim_max[k]);

					p.X = XMAP(j, p.X);
					p.Y = YMAP(k, p.Y);
					q.X = XMAP(j, q.X);
					q.Y = YMAP(k, q.Y);

					canvas->FillRectangle(p.X, p.Y, q.X-p.X, q.Y-p.Y);
				}
			}

			// draw boundary coverage if necessary
			if (brush->bound_type != Brush::STEP)
				DrawScatterBrushBoundRamp(canvas,brush,interface_color[BRUSH1+br],interface_color[BG]);
		}
	}
	
	//jing draw the grid of the plots
    // draw grid
    canvas->SetForeground(interface_color[GRID1]);
    canvas->SetXorMode(false);

    float boxwidth = 1.0/dataset->dims;
    float boxheight = 1.0/dataset->dims;

    int i;
    for (i = 0;i < dataset->dims;i++)
        for (int j = 0;j < dataset->dims;j++)
            canvas->DrawRectangle(i*boxwidth, j*boxheight, boxwidth, boxheight);
//jing draw the labels of the plots
    // do labels
    canvas->SetForeground(interface_color[TEXT]);

    Vec2 font_size = canvas->GetDataFontSize("X");
    float max_str_size = boxwidth - font_size.X;

    // draw horizontal labels
    // add letters one at a time until space runs out
    for (i=0; i<dataset->dims; i++)
    {
        float x = i*boxwidth;
        float y = .8*font_size.Y;
        canvas->DrawConfinedString(x,y,dataset->names[i],max_str_size);
    }

    // draw vertical labels
    // add letters one at a time until space runs out
    for (i=0; i<dataset->dims; i++)
    {
        // initial coordinates
        Vec2 p;
        p.X = -1.5*font_size.X;
        p.Y = i*boxheight; // - font_size.Y*1.1;

        for (int j=0; j<int(strlen(dataset->names[i])); j++)
        {
            Vec2 letter = canvas->GetDataFontSize(&dataset->names[i][j],1);
            p.Y -= 1.2*letter.Y;

            if (p.Y > (i+1)*boxheight+letter.Y) break;

            canvas->DrawString(p.X,p.Y,&dataset->names[i][j],1);
        }
    }

//jing draw points( regular and marked )
    canvas->SetXorMode(false);

    ARR<float> data(dataset->dims);

    // first draw all points in the data color
    //
    canvas->SetForeground(interface_color[DATA]);
    glPointSize((DSIZE+1)*canvas->zoom.X);
    canvas->BeginDraw();
    glBegin(GL_POINTS);

    if ( maskUnbrushedData == 0 )
	{
		for (i=0; i<dataset->data_size; i++)
		{
			// First time through only paint regular data
			//
			if (dataset->mark[i] != okcdata::REGULAR) continue;

			dataset->GetData(data, i);

			for (int j=0; j<dataset->dims; j++) {
				for (int k=0; k<dataset->dims; k++)
					glVertex2f( XMAP(j,data[j]), YMAP(k,data[k]) );
			}
		}
	}
	// now paint all points covered by an operation
    if ( maskBrushedData == 0 )
	{
		for (i=0; i<dataset->data_size; i++)
		{
			if (dataset->mark[i] == okcdata::DISABLED) continue;
			if (dataset->mark[i] == okcdata::REGULAR) continue;

			dataset->GetData(data, i);

			glColor3f(dataset->color[i].R,dataset->color[i].G,dataset->color[i].B);

			for (int j=0; j<dataset->dims; j++) {
				for (int k=0; k<dataset->dims; k++)
					glVertex2f( XMAP(j,data[j]), YMAP(k,data[k]) );
			}
		}
	}
    glEnd();
    canvas->EndDraw();

//jing no use
    // Draw average operations
    for (int op=0; op<num_oper; op++)
    {
        // Check if this operation has OPER_AVERAGE set */
        if (!(brush_ops[op].oper & AVERAGE)) continue;

        ARR<float> average(dataset->dims);
        if (BrushOperCompAverage(op, average)==0) continue;

//      glLineWidth(3.0);
        canvas->SetForeground(interface_color[TEXT]);
        DrawScatterAverage(canvas,average);
    }
    glPointSize(1.0);
}


//
// Set of static variables to monitor
// current state of interaction
//
static bool xor_needs_undo = false;
static int dimx, dimy;
static bool resize_minx, resize_maxx, resize_miny, resize_maxy;
enum INPUT_MODE { RESIZE_BRUSH, MOVE_BRUSH, RESIZE_RAMP, MULTI_SELECT, CANVAS_PAN, NONE };
static INPUT_MODE input_mode = NONE;

//jing this func is used by canvas->.._proc
static void InputScatterResizeBrush(grCanvas *canvas, int sx, int sy)
{
    Vec2 p = canvas->MapScreenToData( Vec2(sx,sy) );

    // Map mouse x/y location to data coordinates
    float mx = XINVMAP(dimx, p.X);
    float my = YINVMAP(dimy, p.Y);

    // If this is a diagonal plot (same dimension on both axes)
    // then average the x and y positions
    //
    if (dimx == dimy) mx = my = .5*(mx+my);

    // Bound mouse to plot edges
    //
    mx = MIN(mx, dataset->dim_max[dimx]);
    mx = MAX(mx, dataset->dim_min[dimx]);
    my = MIN(my, dataset->dim_max[dimy]);
    my = MAX(my, dataset->dim_min[dimy]);

    // Check to make sure min<max
    if (resize_minx && mx >= curr_brush->Max(dimx))
        mx = curr_brush->Max(dimx);
    if (resize_maxx && mx <= curr_brush->Min(dimx))
        mx = curr_brush->Min(dimx);
    if (resize_miny && my >= curr_brush->Max(dimy))
        my = curr_brush->Max(dimy);
    if (resize_maxy && my <= curr_brush->Min(dimy))
        my = curr_brush->Min(dimy);

    // Find the new bounds based on what is being resized
    //
    float new_minx, new_maxx, new_miny, new_maxy;

    if (resize_minx) new_minx = mx;
    else new_minx = curr_brush->Min(dimx);

    if (resize_maxx) new_maxx = mx;
    else new_maxx = curr_brush->Max(dimx);

    if (resize_miny) new_miny = my;
    else new_miny = curr_brush->Min(dimy);

    if (resize_maxy) new_maxy = my;
    else new_maxy = curr_brush->Max(dimy);

    if (resize_minx || resize_maxx)
    {
        curr_brush->SetMinMax(dimx,new_minx,new_maxx);
    }
    if (resize_miny || resize_maxy)
    {
        curr_brush->SetMinMax(dimy,new_miny,new_maxy);
    }

    canvas->SetXorMode(true);

    // Do the graph being manipulated
    DrawScatterBrushBox(canvas, dimx, dimy,
                  (new_maxx + new_minx)/2,
                  (new_maxy + new_miny)/2,
                  (new_maxx - new_minx),
                  (new_maxy - new_miny));

    // Do all other graphs along the horizontal
    for (int x=0; x<dataset->dims; x++)
    {
        if (x==dimx) continue;

        DrawScatterBrushBox(canvas, x, dimy,
                  curr_brush->pos[x],
                  (new_maxy + new_miny)/2,
                  curr_brush->size[x],
                  (new_maxy - new_miny));
    }

    // Do all other graphs along the vertical
    for (int y=0; y<dataset->dims; y++)
    {
        if (y==dimy) continue;

        DrawScatterBrushBox(canvas, dimx, y,
                  (new_maxx + new_minx)/2,
                  curr_brush->pos[y],
                  (new_maxx - new_minx),
                  curr_brush->size[y]);
    }

    canvas->SetXorMode(false);
    xor_needs_undo = true;
}

//jing this func is used by canvas->.._proc
static void InputScatterResizeBoundRamp(grCanvas *canvas, int sx, int sy)
{
    Vec2 p = canvas->MapScreenToData( Vec2(sx,sy) );

    // Map mouse x/y location to data coordinates
    float mx = XINVMAP(dimx, p.X);
    float my = YINVMAP(dimy, p.Y);

    // If this is a diagonal plot (same dimension on both axes)
    // then average the x and y positions
    //
    if (dimx == dimy) mx = my = .5*(mx+my);

    // Bound mouse to plot edges
    //
    mx = MIN(mx, dataset->dim_max[dimx]);
    mx = MAX(mx, dataset->dim_min[dimx]);
    my = MIN(my, dataset->dim_max[dimy]);
    my = MAX(my, dataset->dim_min[dimy]);

    // Check to make sure min<max
    if (resize_minx && mx >= curr_brush->Max(dimx))
        mx = curr_brush->Max(dimx);
    if (resize_maxx && mx <= curr_brush->Min(dimx))
        mx = curr_brush->Min(dimx);
    if (resize_miny && my >= curr_brush->Max(dimy))
        my = curr_brush->Max(dimy);
    if (resize_maxy && my <= curr_brush->Min(dimy))
        my = curr_brush->Min(dimy);

    if (resize_minx)
    {
        if (mx > curr_brush->Min(dimx))
            curr_brush->bound_offset[dimx] = 0.0;
        else
            curr_brush->bound_offset[dimx] = curr_brush->Min(dimx) - mx;
    }

    if (resize_maxx)
    {
        if (mx < curr_brush->Max(dimx))
            curr_brush->bound_offset[dimx] = 0.0;
        else
            curr_brush->bound_offset[dimx] = mx - curr_brush->Max(dimx);
    }

    if (resize_miny)
    {
        if (my > curr_brush->Min(dimy))
            curr_brush->bound_offset[dimy] = 0.0;
        else
            curr_brush->bound_offset[dimy] = curr_brush->Min(dimy) - my;
    }

    if (resize_maxy)
    {
        if (my < curr_brush->Max(dimy))
            curr_brush->bound_offset[dimy] = 0.0;
        else
            curr_brush->bound_offset[dimy] = my - curr_brush->Max(dimy);
    }

    DrawScatterBrushBoundRampOutline(canvas,curr_brush);

    xor_needs_undo = true;
}

//jing this func is used by canvas->.. _proc
static void InputScatterMoveBrush(grCanvas *canvas, int sx, int sy)
{
    Vec2 p = canvas->MapScreenToData( Vec2(sx,sy) );

    // Map mouse x/y location to data coordinates
    float mx = XINVMAP(dimx, p.X);
    float my = YINVMAP(dimy, p.Y);

    // If this is a diagonal plot (same dimension on both axes)
    // then average the x and y positions
    //
    if (dimx == dimy) mx = my = .5*(mx+my);

    // Bound mouse to plot edges
    //
    mx = MIN(mx, dataset->dim_max[dimx] - curr_brush->size[dimx]/2);
    mx = MAX(mx, dataset->dim_min[dimx] + curr_brush->size[dimx]/2);

    my = MIN(my, dataset->dim_max[dimy] - curr_brush->size[dimy]/2);
    my = MAX(my, dataset->dim_min[dimy] + curr_brush->size[dimy]/2);

    curr_brush->pos[dimx] = mx;
    curr_brush->pos[dimy] = my;

    canvas->SetXorMode(true);

    DrawScatterBrushBox(canvas, dimx, dimy, mx, my, curr_brush->size[dimx], curr_brush->size[dimy]);

    // Do all other graphs along the horizontal
    for (int x=0; x<dataset->dims; x++)
    {
        if (x==dimx) continue;

        DrawScatterBrushBox(canvas, x, dimy, curr_brush->pos[x], my,
            curr_brush->size[x], curr_brush->size[dimy]);
    }

    // Do all other graphs along the vertical
    for (int y=0; y<dataset->dims; y++)
    {
        if (y==dimy) continue;

        DrawScatterBrushBox(canvas, dimx, y, mx, curr_brush->pos[y],
            curr_brush->size[dimx], curr_brush->size[y]);
    }

    canvas->SetXorMode(false);
    xor_needs_undo = true;
}

//jing this func is used by canvas->.. _proc
static void InputScatterMultiSelect(grCanvas *canvas, int sx, int sy)
{
    if (!multi_select) return;

    ARR<float> data(dataset->dims);

    // Find all points within circle distance of mouse
    //
    for (int i=0; i<dataset->data_size; i++)
    {
        dataset->GetData(data, i);

        Vec2 p;
        p.X = XMAP(dimx,data[dimx]);
        p.Y = YMAP(dimy,data[dimy]);
        Vec2 q = canvas->MapDataToScreen(p);

        const int rad2 = multisel_radius*multisel_radius; // radius^2 in screen coords

        if ((q.X-sx)*(q.X-sx) + (q.Y-sy)*(q.Y-sy) < rad2)
        {
            multiSelectAddPoint(i);
            glDrawBuffer(GL_FRONT);
            canvas->SetForeground(interface_color[PAINT]);
            RedrawSingleScatter(canvas, i);
            glDrawBuffer(GL_BACK);
        }
    }
}

//jing this func is used by canvas->.. _proc
static void ScatterCaption(grCanvas *canvas, int x, int y)
{
    Vec2 p = canvas->MapScreenToData( Vec2(x,y) );

    float fx = p.X * dataset->dims;
    float fy = p.Y * dataset->dims;
    if (fx<0 || fy<0) return;

    int dim_x = int(fx);
    int dim_y = int(fy);
    dim_x = CLAMP(dim_x,0,dataset->dims-1);
    dim_y = CLAMP(dim_y,0,dataset->dims-1);

    float datax = XINVMAP(dim_x,p.X);
    float datay = YINVMAP(dim_y,p.Y);

    char caption[100];
    sprintf(caption,"%s=",dataset->names[dim_x]);

    if (dataset->nominal_fields[dim_x] != 0) {
//        int value = (int) rint(datax);
//jing 
		int value = (int)(datax+0.5);
//jing
        const char *string = dataset->nominal_fields[dim_x]->MapValueToString(value);
        strcat(caption,string);
        strcat(caption," ");
    } else {
        sprintf(caption, "%s%.2f ", caption, datax);
    }

    strcat(caption,", ");
    strcat(caption,dataset->names[dim_y]);
    strcat(caption,"=");

    if (dataset->nominal_fields[dim_y] != 0) {
 //       int value = (int) rint(datay);
//jing
		int value = (int)(datay+0.5);
//jing

        const char *string = dataset->nominal_fields[dim_y]->MapValueToString(value);
        strcat(caption,string);
        strcat(caption," ");
    } else {
        sprintf(caption, "%s%.2f ", caption, datay);
    }

    appShowCaption(Togl_Interp(canvas->GetWidget()),caption);
}

//jing canvas->motion_proc, Rewrite in hier
void InputScatterMouseMotion(grCanvas *canvas, int x, int y)
{
    if (!dataset) return;

    ScatterCaption(canvas, x, y);
    static int prev_sx, prev_sy;

    switch (input_mode) {
        case MULTI_SELECT:
            InputScatterMultiSelect(canvas,x,y);
            return;

        case RESIZE_BRUSH:
            if (xor_needs_undo) InputScatterResizeBrush(canvas,prev_sx,prev_sy);
            InputScatterResizeBrush(canvas,x,y);
            break;

        case MOVE_BRUSH:
            if (xor_needs_undo) InputScatterMoveBrush(canvas,prev_sx,prev_sy);
            InputScatterMoveBrush(canvas,x,y);
            break;

        case RESIZE_RAMP:
            if (xor_needs_undo) InputScatterResizeBoundRamp(canvas,prev_sx,prev_sy);
            InputScatterResizeBoundRamp(canvas,x,y);
            break;

        case CANVAS_PAN:
            appCanvasMousePan(canvas);
            break;

        default: break;
    }
    prev_sx = x;
    prev_sy = y;
}

//jing this func is used by canvas->.. _proc
static void FindResizeCorners(float x,float y)
{
    // There are four different variables that could be resized
    // (minx, maxx, miny, maxy) the user can resize at most
    // two at once

    // Reset all resize flags
    resize_minx = resize_maxx = false;
    resize_miny = resize_maxy = false;

    // Map mouse x/y location to data coordinates
    float mx = XINVMAP(dimx, x);
    float my = YINVMAP(dimy, y);

    // If this is a diagonal plot (same dimension on both axes)
    // then average the x and y positions
    //
    if (dimx == dimy) mx = my = .5*(mx+my);

    // Check if user is resizing in x direction
    if (mx <= curr_brush->pos[dimx] - .25*curr_brush->size[dimx])
        resize_minx = true;
    else
    if (mx >= curr_brush->pos[dimx] + .25*curr_brush->size[dimx])
        resize_maxx = true;

    // Check if user is resizing in y direction
    if (my <= curr_brush->pos[dimy] - .25*curr_brush->size[dimy])
        resize_miny = true;
    else
    if (my >= curr_brush->pos[dimy] + .25*curr_brush->size[dimy])
        resize_maxy = true;
}

//jing canvas->buttonpress_proc, Rewrite in hier
void InputScatterButtonPress(grCanvas *canvas, int button, int x, int y)
{
    if (!dataset) return;

    if (canvas->button[2]==true) {
        input_mode = CANVAS_PAN;
        canvas->SetHandCursor();
        return;
    }

    Vec2 p = canvas->MapScreenToData( Vec2(x,y) );

    float fx = p.X * dataset->dims;
    float fy = p.Y * dataset->dims;
    if (fx<0 || fy<0) return;

    dimx = int(fx);
    dimy = int(fy);

    if (dimx < 0 || dimx >= dataset->dims || dimy < 0 || dimy >= dataset->dims) {
        dimx = dimy = -1;
        return;
    }

    if (canvas->shiftkey) {
        input_mode = MULTI_SELECT;
        InputScatterMultiSelect(canvas,x,y);
        return;
    } else
    if (canvas->controlkey && curr_brush->bound_type == Brush::RAMP && button == 0) {
        input_mode = RESIZE_RAMP;
        FindResizeCorners(p.X,p.Y);
        InputScatterResizeBoundRamp(canvas,x,y);
        return;
    }

    assert(xor_needs_undo==false);

    if (canvas->button[0] == true) {
        input_mode = RESIZE_BRUSH;
        FindResizeCorners(p.X,p.Y);
    } else
    if (canvas->button[1] == true) {
        input_mode = MOVE_BRUSH;
    }

    InputScatterMouseMotion(canvas,x,y);
}

//jing canvas->buttonrelease_proc, Rewrite in hier
void InputScatterButtonRelease(grCanvas *canvas, int button, int x, int y)
{
    if (!dataset) return;

    switch (input_mode) {
        case MULTI_SELECT:
            break;

        case CANVAS_PAN:
            canvas->SetDefaultCursor();
            break;

        case RESIZE_RAMP:
        case RESIZE_BRUSH:
        case MOVE_BRUSH:
            canvas->SetDefaultCursor();
            BrushOperMarkDataset();
            appRefreshAllCanvas(Togl_Interp(canvas->GetWidget()));
            break;

        default: break;
    }

    input_mode = NONE;
    xor_needs_undo = false;
}
