////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty. 
//  In no event will the authors be held liable for any damages arising 
//  from the use of this software.
// 
//  Authors:
//      Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //
 
#include "stacktool.H"
#include "XmdvTool.H"
#include "dimstack.H"
#include "brush.H"
#include "ogl.H"

const float B = 0.1;
static int odd, even;

void RegisterStackToolCanvas(grCanvas *canvas)
{
    assert(canvas);
    canvas->display_proc = StackBToolRedrawCanvas;
    canvas->motion_proc = StackBToolMotionNotify;
    canvas->buttonpress_proc = StackBToolButtonPress;
    canvas->buttonrelease_proc = StackBToolButtonRelease;
}

static void SetupTopHalfCanvas(grCanvas *canvas)
{
    int W, H;
    canvas->GetSize(&W,&H);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glViewport( 0, H/2, W, H/2);
    glOrtho(-3*B,1+B,-B,1+B,-1,1);
}

static void SetupBottomHalfCanvas(grCanvas *canvas)
{
    int W, H;
    canvas->GetSize(&W,&H);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glViewport( 0, 0, W, H/2);
    glOrtho(-B,1+B,-3*B,1+B,-1,1);
}

void StackBToolRedrawCanvas(grCanvas *canvas)
{
    if (!dataset) return;
    
    canvas->SetBackground(interface_color[BG]);
    canvas->Clear();

    odd = dataset->dims/2 + (dataset->dims%2);
    even = dataset->dims/2;

    // Top Half
    //
    SetupTopHalfCanvas(canvas);

    float interval = 1./odd;
    float thickness = .75*interval;

    int i;
    for (i=0; i<dataset->dims; i++)
    {
        // ODD : Horizontal bars
        if (i%2) continue;

        float lo = (curr_brush->Min(i) - dataset->org_dim_min[i]) /
                (dataset->org_dim_max[i] - dataset->org_dim_min[i]);
        lo = MAX(lo,0);

        float hi = (curr_brush->Max(i) - dataset->org_dim_min[i]) /
                (dataset->org_dim_max[i] - dataset->org_dim_min[i]);
        hi = MIN(hi,1);

        canvas->SetForeground(interface_color[BRUSH1+curr_brush->num]);
        canvas->FillRectangle( lo, (i/2)*interval, hi-lo, thickness);

        canvas->SetForeground(interface_color[GRID1]);
        canvas->DrawRectangle( 0, (i/2)*interval, 1, thickness);

        canvas->SetForeground(interface_color[TEXT]);
        canvas->DrawConfinedString(-2*B, (i/2+.25)*interval, dataset->names[i], 1.9*B);
    }

    // Bottom Half
    //
    SetupBottomHalfCanvas(canvas);
    
    interval = 1./even;
    thickness = .75*interval;

    for (i=0; i<dataset->dims; i++)
    {
        // EVEN
        if (!(i%2)) continue;

        float lo = (curr_brush->Min(i) - dataset->org_dim_min[i]) /
                (dataset->org_dim_max[i] - dataset->org_dim_min[i]);
        lo = MAX(lo,0);

        float hi = (curr_brush->Max(i) - dataset->org_dim_min[i]) /
                (dataset->org_dim_max[i] - dataset->org_dim_min[i]);
        hi = MIN(hi,1);

        canvas->SetForeground(interface_color[BRUSH1+curr_brush->num]);
        canvas->FillRectangle( (i/2)*interval, lo, thickness, hi-lo);

        canvas->SetForeground(interface_color[GRID1]);
        canvas->DrawRectangle( (i/2)*interval, 0, thickness, 1);

        canvas->SetForeground(interface_color[TEXT]);
        canvas->DrawConfinedString( (i/2)*interval, -1.5*B, dataset->names[i], thickness);
    }
}

static int axis = -1;
static int resize_which = -1;
static int top_bottom = -1;

static void StackBToolResize(grCanvas *canvas, int x, int y)
{
    float val;
    
    switch (top_bottom) {
    case 1: {
        SetupBottomHalfCanvas(canvas);
        Vec2 p = canvas->MapScreenToData(Vec2(x,y));
        val = p.Y;
        } break;

    case 0: {
        SetupTopHalfCanvas(canvas);
        Vec2 p = canvas->MapScreenToData(Vec2(x,y));
        val = p.X;

        } break;
    }

    // Calculate new position and bound it 
    float new_pos = dataset->org_dim_min[axis]+val*(dataset->org_dim_max[axis]-dataset->dim_min[axis]);
    new_pos = MIN(new_pos, dataset->org_dim_max[axis]);
    new_pos = MAX(new_pos, dataset->org_dim_min[axis]);

    float lo = curr_brush->Min(axis);
    float hi = curr_brush->Max(axis);

    if (resize_which<0) {
        if (fabs(new_pos-lo) < fabs(new_pos-hi)) {
            resize_which = 0;
        } else {
            resize_which = 1;
        }
    }

    if (resize_which==0) lo = MIN(new_pos,hi);
    else hi = MAX(new_pos, lo);

    curr_brush->SetMinMax(axis,lo,hi);
}

static void StackBToolMove(grCanvas *canvas, int x, int y)
{
    float val;
    
    switch (top_bottom) {
    case 1: {
        SetupBottomHalfCanvas(canvas);
        Vec2 p = canvas->MapScreenToData(Vec2(x,y));
        val = p.Y;
        } break;

    case 0: {
        SetupTopHalfCanvas(canvas);
        Vec2 p = canvas->MapScreenToData(Vec2(x,y));
        val = p.X;

        } break;
    }

    float new_pos = dataset->org_dim_min[axis]+val*(dataset->org_dim_max[axis]-dataset->org_dim_min[axis]); 
    new_pos = MIN(new_pos, dataset->org_dim_max[axis]-curr_brush->size[axis]/2.);
    new_pos = MAX(new_pos, dataset->org_dim_min[axis]+curr_brush->size[axis]/2.);

    curr_brush->pos[axis] = new_pos;

}

void StackBToolMotionNotify(grCanvas *canvas, int x, int y) 
{
    if (!dataset) return;
    if (axis < 0) return;

    if (canvas->button[0]==true) {
        StackBToolResize(canvas,x,y);
        canvas->Refresh();
    } else
    if (canvas->button[1]==true) {
        StackBToolMove(canvas,x,y);
        canvas->Refresh();
    }
}

void StackBToolButtonPress(grCanvas *canvas, int button, int x, int y)
{
    if (!dataset) return;
    
    int W, H;
    canvas->GetSize(&W,&H);
        
    if (y<H/2) {
        top_bottom = 1;
        SetupBottomHalfCanvas(canvas);

        Vec2 p = canvas->MapScreenToData(Vec2(x,y));
        
        float interval = 1./even;

        axis = 2*int(p.X / interval) + 1;

    } else {
        top_bottom = 0;
        SetupTopHalfCanvas(canvas);

        Vec2 p = canvas->MapScreenToData(Vec2(x,y));

        float interval = 1./odd;

        axis = 2*int(p.Y / interval);
    }

    // Sanity check 
    if (axis < 0 || axis >= dataset->dims) {
        axis = -1;
        return;
    }

    StackBToolMotionNotify(canvas,x,y);
}

void StackBToolButtonRelease(grCanvas *canvas, int button, int x, int y)
{
    if (!dataset) return;
    if (axis<0) return;

    appCanvasBrushHasChanged(Togl_Interp(canvas->GetWidget()));
    appRefreshAllCanvas(Togl_Interp(canvas->GetWidget()));
    axis = -1;
    resize_which = -1;
}
